r"""Implement a result that combines a mapping of result objects into a
single result object."""

from __future__ import annotations

__all__ = ["MappingResult"]

from typing import TYPE_CHECKING, Any

from coola import objects_are_equal
from coola.utils import str_indent, str_mapping

from arkas.result import BaseResult

if TYPE_CHECKING:
    from collections.abc import Mapping


class MappingResult(BaseResult):
    r"""Implement a result that combines a mapping of result objects into
    a single result object.

    Args:
        results: The mapping of result objects to combine.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.result import MappingResult, Result
    >>> result = MappingResult(
    ...     {
    ...         "class1": Result(metrics={"accuracy": 62.0, "count": 42}),
    ...         "class2": Result(metrics={"accuracy": 42.0, "count": 42}),
    ...     }
    ... )
    >>> result
    MappingResult(count=2)
    >>> result.compute_metrics()
    {'class1': {'accuracy': 62.0, 'count': 42}, 'class2': {'accuracy': 42.0, 'count': 42}}

    ```
    """

    def __init__(self, results: Mapping[str, BaseResult]) -> None:
        self._results = results

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(count={len(self._results):,})"

    def __str__(self) -> str:
        args = str_indent(str_mapping(self._results))
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def compute_metrics(self, prefix: str = "", suffix: str = "") -> dict:
        return {
            key: result.compute_metrics(prefix=prefix, suffix=suffix)
            for key, result in self._results.items()
        }

    def equal(self, other: Any, equal_nan: bool = False) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return objects_are_equal(self._results, other._results, equal_nan=equal_nan)

    def generate_figures(self, prefix: str = "", suffix: str = "") -> dict:
        return {
            key: result.generate_figures(prefix=prefix, suffix=suffix)
            for key, result in self._results.items()
        }



r"""Contain the implementation of a simple HTML content generator."""

from __future__ import annotations

__all__ = ["BaseSectionContentGenerator", "create_template"]

import logging
from abc import abstractmethod
from typing import TYPE_CHECKING

from jinja2 import Template

from arkas.content.base import BaseContentGenerator
from arkas.utils.html import GO_TO_TOP, render_toc, tags2id, tags2title, valid_h_tag

if TYPE_CHECKING:
    from collections.abc import Sequence

    from arkas.content.vanilla import ContentGenerator

logger = logging.getLogger(__name__)


class BaseSectionContentGenerator(BaseContentGenerator):
    r"""Define a base class to implement a section content generator.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.content import ContentGenerator
    >>> generator = ContentGenerator("meow")
    >>> generator
    ContentGenerator()
    >>> generator.generate_content()
    'meow'

    ```
    """

    def compute(self) -> ContentGenerator:
        # local import to avoid cyclic dependency because ContentGenerator
        # uses this class as base class
        from arkas.content.vanilla import ContentGenerator

        return ContentGenerator(self.generate_content())

    def generate_body(self, number: str = "", tags: Sequence[str] = (), depth: int = 0) -> str:
        return Template(create_template()).render(
            {
                "go_to_top": GO_TO_TOP,
                "id": tags2id(tags),
                "depth": valid_h_tag(depth + 1),
                "title": tags2title(tags),
                "section": number,
                "content": self.generate_content(),
            }
        )

    def generate_toc(
        self, number: str = "", tags: Sequence[str] = (), depth: int = 0, max_depth: int = 1
    ) -> str:
        return render_toc(number=number, tags=tags, depth=depth, max_depth=max_depth)

    @abstractmethod
    def generate_content(self) -> str:
        r"""Return the section content without the tags.

        Returns:
            The content  without the tags.
        """


def create_template() -> str:
    r"""Return the template of the section.

    Returns:
        The section template.

    Example usage:

    ```pycon

    >>> from arkas.content.accuracy import create_template
    >>> template = create_template()

    ```
    """
    return """<h{{depth}} id="{{id}}">{{section}} {{title}} </h{{depth}}>
{{go_to_top}}
<p style="margin-top: 1rem;">
{{content}}
<p style="margin-top: 1rem;">
"""



r"""Contain states."""

from __future__ import annotations

__all__ = ["AccuracyState", "BaseState", "PrecisionRecallState"]

from arkas.state.accuracy import AccuracyState
from arkas.state.base import BaseState
from arkas.state.precision_recall import PrecisionRecallState



r"""Contain HTML content generators."""

from __future__ import annotations

__all__ = [
    "AccuracyContentGenerator",
    "BalancedAccuracyContentGenerator",
    "BaseContentGenerator",
    "ColumnCooccurrenceContentGenerator",
    "ContentGenerator",
    "ContentGeneratorDict",
    "DataFrameSummaryContentGenerator",
]

from arkas.content.accuracy import AccuracyContentGenerator
from arkas.content.balanced_accuracy import BalancedAccuracyContentGenerator
from arkas.content.base import BaseContentGenerator
from arkas.content.column_cooccurrence import ColumnCooccurrenceContentGenerator
from arkas.content.frame_summary import DataFrameSummaryContentGenerator
from arkas.content.mapping import ContentGeneratorDict
from arkas.content.vanilla import ContentGenerator



r"""Contain DataFrame utility functions."""

from __future__ import annotations

__all__ = ["to_arrays"]


from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import numpy as np
    import polars as pl


def to_arrays(frame: pl.DataFrame) -> dict[str, np.ndarray]:
    r"""Convert a ``polars.DataFrame`` to a dictionary of NumPy arrays.

    Args:
        frame: The DataFrame to convert.

    Returns:
        A dictionary of NumPy arrays.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from arkas.utils.dataframe import to_arrays
    >>> frame = pl.DataFrame(
    ...     {
    ...         "int": [1, 2, 3, 4, 5],
    ...         "float": [5.0, 4.0, 3.0, 2.0, 1.0],
    ...         "str": ["a", "b", "c", "d", "e"],
    ...     },
    ...     schema={"int": pl.Int64, "float": pl.Float64, "str": pl.String},
    ... )
    >>> data = to_arrays(frame)
    >>> data
    {'int': array([1, 2, 3, 4, 5]),
     'float': array([5., 4., 3., 2., 1.]),
     'str': array(['a', 'b', 'c', 'd', 'e'], dtype=object)}

    ```
    """
    return {s.name: s.to_numpy() for s in frame.iter_columns()}



r"""Contain figures."""

from __future__ import annotations

__all__ = [
    "BaseFigure",
    "BaseFigureConfig",
    "DefaultFigureConfig",
    "HtmlFigure",
    "MatplotlibFigure",
    "MatplotlibFigureConfig",
    "PlotlyFigure",
    "PlotlyFigureConfig",
    "figure2html",
    "get_default_config",
]

from arkas.figure.base import BaseFigure, BaseFigureConfig
from arkas.figure.default import DefaultFigureConfig
from arkas.figure.html import HtmlFigure
from arkas.figure.matplotlib import MatplotlibFigure, MatplotlibFigureConfig
from arkas.figure.plotly import PlotlyFigure, PlotlyFigureConfig
from arkas.figure.utils import figure2html, get_default_config



r"""Contain some utility functions for command lines."""

from __future__ import annotations

__all__ = ["get_original_cwd", "log_run_info"]

import logging
from pathlib import Path
from unittest.mock import Mock

from coola.utils.path import sanitize_path

import arkas
from arkas.utils.imports import (
    check_omegaconf,
    is_hydra_available,
    is_omegaconf_available,
)

if is_hydra_available():  # pragma: no cover
    import hydra
    from hydra.core.hydra_config import HydraConfig

if is_omegaconf_available():
    from omegaconf import DictConfig, OmegaConf
else:  # pragma: no cover
    DictConfig = Mock()

logger = logging.getLogger(__name__)

# ASCII logo generated from https://ascii.co.uk/art/bear and
# https://patorjk.com/software/taag/#p=testall&f=Graffiti&t=arkas%0A
LOGO = rf"""
           .--.              .--.
          : (\ ". _......_ ." /) :
           '.    `        `    .'                    __
            /'   _        _   `\       _____ _______|  | _______    ______
           /     0}}      {{0     \      \__  \\_  __ \  |/ /\__  \  /  ___/
          |       /      \       |      / __ \|  | \/    <  / __ \_\___ \
          |     /'        `\     |     (____  /__|  |__|_ \(____  /____  >
           \   | .  .==.  . |   /           \/           \/     \/     \/
            '._ \.' \__/ './ _.'
            /  ``'._-''-_.'``  \
                    `--`

version: {arkas.__version__}"""


def get_original_cwd() -> Path:
    r"""Get the original working directory the experiment was launched
    from.

    The problem is that Hydra change the working directory when the
    application is launched.

    Returns:
        If Hydra is initialized, it returns the original working
            directory otherwise it returns the current working
            directory.

    Example usage:

    ```pycon

    >>> from arkas.cli.utils import get_original_cwd
    >>> get_original_cwd()

    ```
    """
    if is_hydra_available() and HydraConfig.initialized():
        return sanitize_path(hydra.utils.get_original_cwd())
    return Path.cwd()


def log_run_info(config: DictConfig) -> None:
    """Log some information about the current run.

    Args:
        config: The config of the run.

    Example usage:

    ```pycon

    >>> from omegaconf import OmegaConf
    >>> from arkas.cli.utils import log_run_info
    >>> log_run_info(OmegaConf.create())

    ```
    """
    check_omegaconf()
    logger.info(LOGO)
    logger.info("Original working directory: %s", get_original_cwd())
    logger.info("Current working directory: %s", Path.cwd())
    logger.info("Config:\n%s", OmegaConf.to_yaml(config))



r"""Contain precision-recall curve plotting functionalities."""

from __future__ import annotations

__all__ = ["binary_precision_recall_curve"]

from typing import TYPE_CHECKING, Any

from sklearn.metrics import PrecisionRecallDisplay

from arkas.metric.utils import preprocess_pred

if TYPE_CHECKING:
    import numpy as np
    from matplotlib.axes import Axes


def binary_precision_recall_curve(
    ax: Axes, y_true: np.ndarray, y_pred: np.ndarray, **kwargs: Any
) -> None:
    r"""Plot the precision-recall curve for binary labels.

    Args:
        ax: The axes of the matplotlib figure to update.
        y_true: The ground truth target labels. This input must
            be an array of shape ``(n_samples,)`` with ``0`` and
            ``1`` values.
        y_pred: The predicted labels. This input must be an array of
            shape ``(n_samples,)`` with ``0`` and ``1`` values.
        **kwargs: Arbitrary keyword arguments that are passed to
            ``PrecisionRecallDisplay.from_predictions``.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from matplotlib import pyplot as plt
    >>> from arkas.plot import binary_precision_recall_curve
    >>> fig, ax = plt.subplots()
    >>> binary_precision_recall_curve(
    ...     ax=ax, y_true=np.array([1, 0, 0, 1, 1]), y_pred=np.array([1, 0, 0, 1, 1])
    ... )

    ```
    """
    y_true, y_pred = preprocess_pred(y_true=y_true.ravel(), y_pred=y_pred.ravel(), drop_nan=True)
    PrecisionRecallDisplay.from_predictions(y_true=y_true, y_pred=y_pred, ax=ax, **kwargs)



r"""Contain the base class to implement an evaluator."""

from __future__ import annotations

__all__ = ["BaseEvaluator", "is_evaluator_config", "setup_evaluator"]

import logging
from abc import ABC
from typing import TYPE_CHECKING

from objectory import AbstractFactory
from objectory.utils import is_object_config

if TYPE_CHECKING:
    import polars as pl

    from arkas.result import BaseResult

logger = logging.getLogger(__name__)


class BaseEvaluator(ABC, metaclass=AbstractFactory):
    r"""Define the base class to evaluate a DataFrame.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from arkas.evaluator import AccuracyEvaluator
    >>> evaluator = AccuracyEvaluator(y_true="target", y_pred="pred")
    >>> evaluator
    AccuracyEvaluator(y_true='target', y_pred='pred', drop_nulls=True, nan_policy='propagate')
    >>> data = pl.DataFrame({"pred": [3, 2, 0, 1, 0, 1], "target": [3, 2, 0, 1, 0, 1]})
    >>> result = evaluator.evaluate(data)
    >>> result
    AccuracyResult(y_true=(6,), y_pred=(6,), nan_policy='propagate')

    ```
    """

    def evaluate(self, data: pl.DataFrame, lazy: bool = True) -> BaseResult:
        r"""Evaluate the result.

        Args:
            data: The data to evaluate.
            lazy: If ``True``, it forces the computation of the
                result, otherwise it returns a result object that
                delays the evaluation of the result.

        Returns:
            The generated result.

        Example usage:

        ```pycon

        >>> import polars as pl
        >>> from arkas.evaluator import AccuracyEvaluator
        >>> evaluator = AccuracyEvaluator(y_true="target", y_pred="pred")
        >>> data = pl.DataFrame({"pred": [3, 2, 0, 1, 0, 1], "target": [3, 2, 0, 1, 0, 1]})
        >>> result = evaluator.evaluate(data)
        >>> result
        AccuracyResult(y_true=(6,), y_pred=(6,), nan_policy='propagate')

        ```
        """


def is_evaluator_config(config: dict) -> bool:
    r"""Indicate if the input configuration is a configuration for a
    ``BaseEvaluator``.

    This function only checks if the value of the key  ``_target_``
    is valid. It does not check the other values. If ``_target_``
    indicates a function, the returned type hint is used to check
    the class.

    Args:
        config: The configuration to check.

    Returns:
        ``True`` if the input configuration is a configuration
            for a ``BaseEvaluator`` object.

    Example usage:

    ```pycon

    >>> from arkas.evaluator import is_evaluator_config
    >>> is_evaluator_config({"_target_": "arkas.evaluator.AccuracyEvaluator"})
    True

    ```
    """
    return is_object_config(config, BaseEvaluator)


def setup_evaluator(
    evaluator: BaseEvaluator | dict,
) -> BaseEvaluator:
    r"""Set up an evaluator.

    The evaluator is instantiated from its configuration
    by using the ``BaseEvaluator`` factory function.

    Args:
        evaluator: An evaluator or its configuration.

    Returns:
        An instantiated evaluator.

    Example usage:

    ```pycon

    >>> from arkas.evaluator import setup_evaluator
    >>> evaluator = setup_evaluator(
    ...     {
    ...         "_target_": "arkas.evaluator.AccuracyEvaluator",
    ...         "y_true": "target",
    ...         "y_pred": "pred",
    ...     }
    ... )
    >>> evaluator
    AccuracyEvaluator(y_true='target', y_pred='pred', drop_nulls=True, nan_policy='propagate')

    ```
    """
    if isinstance(evaluator, dict):
        logger.info("Initializing an evaluator from its configuration... ")
        evaluator = BaseEvaluator.factory(**evaluator)
    if not isinstance(evaluator, BaseEvaluator):
        logger.warning(f"evaluator is not a `BaseEvaluator` (received: {type(evaluator)})")
    return evaluator



r"""Implement a output that combines a mapping of output objects into a
single output object."""

from __future__ import annotations

__all__ = ["OutputDict"]

from typing import TYPE_CHECKING, Any

from coola import objects_are_equal
from coola.utils import str_indent, str_mapping

from arkas.content.mapping import ContentGeneratorDict
from arkas.evaluator2.mapping import EvaluatorDict
from arkas.output.lazy import BaseLazyOutput
from arkas.plotter.mapping import PlotterDict

if TYPE_CHECKING:
    from collections.abc import Mapping

    from arkas.output.base import BaseOutput


class OutputDict(BaseLazyOutput):
    r"""Implement an output that combines a mapping of output objects
    into a single output object.

    Args:
        outputs: The mapping of output objects to combine.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.output import OutputDict, Output, AccuracyOutput
    >>> from arkas.content import ContentGenerator
    >>> from arkas.evaluator2 import Evaluator
    >>> from arkas.plotter import Plotter
    >>> from arkas.state import AccuracyState
    >>> output = OutputDict(
    ...     {
    ...         "one": Output(
    ...             content=ContentGenerator("meow"), evaluator=Evaluator(), plotter=Plotter()
    ...         ),
    ...         "two": AccuracyOutput(
    ...             AccuracyState(
    ...                 y_true=np.array([1, 0, 0, 1, 1]),
    ...                 y_pred=np.array([1, 0, 0, 1, 1]),
    ...                 y_true_name="target",
    ...                 y_pred_name="pred",
    ...             )
    ...         ),
    ...     }
    ... )
    >>> output
    OutputDict(count=2)
    >>> output.get_content_generator()
    ContentGeneratorDict(
      (one): ContentGenerator()
      (two): AccuracyContentGenerator(
          (state): AccuracyState(y_true=(5,), y_pred=(5,), y_true_name='target', y_pred_name='pred')
          (nan_policy): propagate
        )
    )
    >>> output.get_evaluator()
    EvaluatorDict(
      (one): Evaluator(count=0)
      (two): AccuracyEvaluator(
          (state): AccuracyState(y_true=(5,), y_pred=(5,), y_true_name='target', y_pred_name='pred')
          (nan_policy): propagate
        )
    )
    >>> output.get_plotter()
    PlotterDict(
      (one): Plotter(count=0)
      (two): Plotter(count=0)
    )

    ```
    """

    def __init__(self, outputs: Mapping[str, BaseOutput]) -> None:
        self._outputs = outputs

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(count={len(self._outputs):,})"

    def __str__(self) -> str:
        args = str_indent(str_mapping(self._outputs))
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def equal(self, other: Any, equal_nan: bool = False) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return objects_are_equal(self._outputs, other._outputs, equal_nan=equal_nan)

    def _get_content_generator(self) -> ContentGeneratorDict:
        return ContentGeneratorDict(
            {key: output.get_content_generator() for key, output in self._outputs.items()}
        )

    def _get_evaluator(self) -> EvaluatorDict:
        return EvaluatorDict({key: output.get_evaluator() for key, output in self._outputs.items()})

    def _get_plotter(self) -> PlotterDict:
        return PlotterDict({key: output.get_plotter() for key, output in self._outputs.items()})



r"""Contain the implementation of a simple reporter that generates a
HTML file and save it."""

from __future__ import annotations

__all__ = ["Reporter"]

import logging
from typing import TYPE_CHECKING

from coola.utils import str_indent, str_mapping
from coola.utils.path import sanitize_path
from iden.io import save_text

from arkas.reporter.base import BaseReporter
from arkas.reporter.utils import create_html_report

if TYPE_CHECKING:
    from pathlib import Path

    from arkas.content import BaseContentGenerator

logger = logging.getLogger(__name__)


class Reporter(BaseReporter):
    r"""Implement a simple reporter that generates a HTML file and save
    it.

    Args:
        report_path: The path where to save the HTML report.
        max_toc_depth: The maximum level to show in the
            table of content.

    Example usage:

    ```pycon

    >>> import tempfile
    >>> from pathlib import Path
    >>> import numpy as np
    >>> from arkas.content import AccuracyContentGenerator
    >>> from arkas.state import AccuracyState
    >>> from arkas.reporter import Reporter
    >>> generator = AccuracyContentGenerator(
    ...     state=AccuracyState(
    ...         y_true=np.array([1, 0, 0, 1, 1]),
    ...         y_pred=np.array([1, 0, 0, 1, 1]),
    ...         y_true_name="target",
    ...         y_pred_name="pred",
    ...     )
    ... )
    >>> with tempfile.TemporaryDirectory() as tmpdir:
    ...     reporter = Reporter(
    ...         report_path=Path(tmpdir).joinpath("report.html"),
    ...     )
    ...     reporter.generate(generator)
    ...

    ```
    """

    def __init__(self, report_path: Path | str, max_toc_depth: int = 6) -> None:
        self._report_path = sanitize_path(report_path)
        self._max_toc_depth = int(max_toc_depth)

    def __repr__(self) -> str:
        args = str_indent(
            str_mapping(
                {
                    "report_path": self._report_path,
                    "max_toc_depth": self._max_toc_depth,
                }
            )
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def generate(self, generator: BaseContentGenerator) -> None:
        logger.info("Creating the HTML report with the results...")
        report = create_html_report(
            toc=generator.generate_body(),
            body=generator.generate_toc(max_depth=self._max_toc_depth),
        )
        logger.info(f"Saving the HTML report at {self._report_path}...")
        save_text(report, self._report_path, exist_ok=True)



r"""Contain reporters."""

from __future__ import annotations

__all__ = ["BaseReporter", "EvalReporter", "Reporter", "is_reporter_config", "setup_reporter"]

from arkas.reporter.base import BaseReporter, is_reporter_config, setup_reporter
from arkas.reporter.eval import EvalReporter
from arkas.reporter.vanilla import Reporter



r"""Contain DataFrame analyzers."""

from __future__ import annotations

__all__ = [
    "AccuracyAnalyzer",
    "BalancedAccuracyAnalyzer",
    "BaseAnalyzer",
    "BaseInNLazyAnalyzer",
    "BaseLazyAnalyzer",
    "BaseTruePredAnalyzer",
    "ColumnCooccurrenceAnalyzer",
    "ContentAnalyzer",
    "DataFrameSummaryAnalyzer",
    "MappingAnalyzer",
    "TransformAnalyzer",
    "is_analyzer_config",
    "setup_analyzer",
]

from arkas.analyzer.accuracy import AccuracyAnalyzer
from arkas.analyzer.balanced_accuracy import BalancedAccuracyAnalyzer
from arkas.analyzer.base import BaseAnalyzer, is_analyzer_config, setup_analyzer
from arkas.analyzer.column_cooccurrence import ColumnCooccurrenceAnalyzer
from arkas.analyzer.columns import BaseTruePredAnalyzer
from arkas.analyzer.content import ContentAnalyzer
from arkas.analyzer.frame_summary import DataFrameSummaryAnalyzer
from arkas.analyzer.lazy import BaseInNLazyAnalyzer, BaseLazyAnalyzer
from arkas.analyzer.mapping import MappingAnalyzer
from arkas.analyzer.transform import TransformAnalyzer



r"""Contain utility functions to manage matplotlib figures."""

from __future__ import annotations

__all__ = ["MISSING_FIGURE_MESSAGE", "figure2html"]

import base64
import io

from matplotlib import pyplot as plt

MISSING_FIGURE_MESSAGE = (
    "<span>&#9888;</span> No figure is generated because of missing or incorrect data"
)


def figure2html(fig: plt.Figure | None, reactive: bool = True, close_fig: bool = False) -> str:
    r"""Convert a matplotlib figure to a string that can be used in a
    HTML file.

    Args:
        fig: The figure to convert.
        reactive: If ``True``, the generated is configured to be
            reactive to the screen size.
        close_fig: If ``True``, the figure is closed after it is
            converted to HTML format.

    Returns:
        The converted figure to a string.

    Example usage:

    ```pycon

    >>> from matplotlib import pyplot as plt
    >>> from arkas.utils.figure import figure2html
    >>> fig, ax = plt.subplots()
    >>> string = figure2html(fig)

    ```
    """
    if fig is None:
        return MISSING_FIGURE_MESSAGE
    fig.tight_layout()
    img = io.BytesIO()
    fig.savefig(img, format="png", bbox_inches="tight")
    img.seek(0)
    data = base64.b64encode(img.getvalue()).decode("utf-8")
    if close_fig:
        plt.close(fig)
    style = 'style="width:100%; height:auto;" ' if reactive else ""
    return f'<img {style}src="data:image/png;charset=utf-8;base64, {data}">'



r"""Define some template classes to implement some analyzers."""

from __future__ import annotations

__all__ = ["BaseTruePredAnalyzer"]

import logging
from abc import abstractmethod
from typing import TYPE_CHECKING

import polars as pl
from grizz.utils.column import check_column_missing_policy, check_missing_column
from grizz.utils.format import str_shape_diff
from polars import selectors as cs

from arkas.analyzer.base import BaseAnalyzer
from arkas.output.empty import EmptyOutput

if TYPE_CHECKING:
    from arkas.output.base import BaseOutput

logger = logging.getLogger(__name__)


class BaseTruePredAnalyzer(BaseAnalyzer):
    r"""Define a base class to implement ``polars.DataFrame``
    analyzer that takes two input columns: ``y_true`` and ``y_pred``.

    Args:
        y_true: The column name of the ground truth target
            labels.
        y_pred: The column name of the predicted labels.
        drop_nulls: If ``True``, the rows with null values in
            ``y_true`` or ``y_pred`` columns are dropped.
        missing_policy: The policy on how to handle missing columns.
            The following options are available: ``'ignore'``,
            ``'warn'``, and ``'raise'``. If ``'raise'``, an exception
            is raised if at least one column is missing.
            If ``'warn'``, a warning is raised if at least one column
            is missing and the missing columns are ignored.
            If ``'ignore'``, the missing columns are ignored and
            no warning message appears.
    """

    def __init__(
        self,
        y_true: str,
        y_pred: str,
        drop_nulls: bool,
        missing_policy: str,
    ) -> None:
        self._y_true = y_true
        self._y_pred = y_pred
        self._drop_nulls = bool(drop_nulls)

        check_column_missing_policy(missing_policy)
        self._missing_policy = missing_policy

    # def __repr__(self) -> str:
    #     args = repr_mapping_line(
    #         {
    #             "y_true": self._y_true,
    #             "y_pred": self._y_pred,
    #             "drop_nulls": self._drop_nulls,
    #             "missing_policy": self._missing_policy,
    #         }
    #     )
    #     return f"{self.__class__.__qualname__}({args})"

    def analyze(self, frame: pl.DataFrame, lazy: bool = True) -> BaseOutput:
        self._check_input_column(frame)
        for col in [self._y_true, self._y_pred]:
            if col not in frame:
                logger.info(
                    f"Skipping '{self.__class__.__qualname__}.analyze' "
                    f"because the input column {col!r} is missing"
                )
                return EmptyOutput()
        output = self._analyze(self._prepare_data(frame))
        if not lazy:
            output = output.compute()
        return output

    def _prepare_data(self, data: pl.DataFrame) -> pl.DataFrame:
        if self._drop_nulls:
            cols = [self._y_true, self._y_pred]
            logger.info(f"Dropping rows that have at least one null value in the columns: {cols}")
            initial_shape = data.shape
            data = data.drop_nulls(cs.by_name(cols))
            logger.info(str_shape_diff(orig=initial_shape, final=data.shape))
        return data

    def _check_input_column(self, frame: pl.DataFrame) -> None:
        r"""Check if the input column is missing.

        Args:
            frame: The input DataFrame to check.
        """
        check_missing_column(frame, column=self._y_true, missing_policy=self._missing_policy)
        check_missing_column(frame, column=self._y_pred, missing_policy=self._missing_policy)

    @abstractmethod
    def _analyze(self, frame: pl.DataFrame) -> BaseOutput:
        r"""Analyze the DataFrame.

        Args:
            frame: The DataFrame to analyze.

        Returns:
            The generated output.
        """



r"""Contain data evaluators."""

from __future__ import annotations

__all__ = [
    "AccuracyEvaluator",
    "BalancedAccuracyEvaluator",
    "BaseEvaluator",
    "Evaluator",
    "EvaluatorDict",
]

from arkas.evaluator2.accuracy import AccuracyEvaluator
from arkas.evaluator2.balanced_accuracy import BalancedAccuracyEvaluator
from arkas.evaluator2.base import BaseEvaluator
from arkas.evaluator2.mapping import EvaluatorDict
from arkas.evaluator2.vanilla import Evaluator



r"""Contain the base class to implement a lazy evaluator."""

from __future__ import annotations

__all__ = ["BaseLazyEvaluator"]

import logging
from abc import abstractmethod
from typing import TYPE_CHECKING, Generic, TypeVar

from grizz.utils.format import str_row_diff
from polars import selectors as cs

from arkas.evaluator.base import BaseEvaluator
from arkas.result import BaseResult, EmptyResult, Result
from arkas.utils.data import find_missing_keys

if TYPE_CHECKING:
    import polars as pl


logger = logging.getLogger(__name__)

T = TypeVar("T", bound=BaseResult)


class BaseLazyEvaluator(BaseEvaluator, Generic[T]):
    r"""Define the base class to evaluate the result lazily.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from arkas.evaluator import AccuracyEvaluator
    >>> evaluator = AccuracyEvaluator(y_true="target", y_pred="pred")
    >>> evaluator
    AccuracyEvaluator(y_true='target', y_pred='pred', drop_nulls=True, nan_policy='propagate')
    >>> data = pl.DataFrame({"pred": [3, 2, 0, 1, 0, 1], "target": [3, 2, 0, 1, 0, 1]})
    >>> result = evaluator.evaluate(data)
    >>> result
    AccuracyResult(y_true=(6,), y_pred=(6,), nan_policy='propagate')

    ```
    """

    def __init__(self, drop_nulls: bool) -> None:
        self._drop_nulls = bool(drop_nulls)

    def _evaluate(self, data: pl.DataFrame, lazy: bool = True) -> T | Result:
        r"""Evaluate the result.

        Args:
            data: The data to evaluate.
            lazy: If ``True``, it forces the computation of the
                result, otherwise it returns a result object that
                delays the evaluation of the result.

        Returns:
            The generated result.
        """
        if missing_cols := find_missing_keys(keys=set(data.columns), queries=self._get_columns()):
            logger.warning(
                "Skipping the evaluation because some columns are missing: "
                f"{sorted(missing_cols)}"
            )
            return EmptyResult()

        data = self._prepare_data(data)
        out = self._compute_result(data)
        if lazy or isinstance(out, EmptyResult):
            return out
        return Result(metrics=out.compute_metrics(), figures=out.generate_figures())

    def _prepare_data(self, data: pl.DataFrame) -> pl.DataFrame:
        if self._drop_nulls:
            cols = self._get_columns()
            logger.info(f"Dropping rows that have at least one null value in the columns: {cols}")
            initial_shape = data.shape
            data = data.drop_nulls(cs.by_name(cols))
            logger.info(
                f"DataFrame shape: {initial_shape} -> {data.shape} | "
                f"{str_row_diff(orig=initial_shape[0], final=data.shape[0])}"
            )
        return data

    @abstractmethod
    def _compute_result(self, data: pl.DataFrame) -> T:
        r"""Compute and return the result.

        Args:
            data: The data to evaluate.

        Returns:
            The generated result.
        """

    @abstractmethod
    def _get_columns(self) -> tuple[str, ...]:
        r"""Get the columns used to compute the result.

        Returns:
            The column names.
        """



r"""Contain utility functions to manage figures."""

from __future__ import annotations

__all__ = ["MISSING_FIGURE_MESSAGE", "figure2html", "get_default_config"]

from typing import TYPE_CHECKING

from arkas.figure.matplotlib import MatplotlibFigureConfig

if TYPE_CHECKING:
    from arkas.figure.base import BaseFigure, BaseFigureConfig

MISSING_FIGURE_MESSAGE = (
    "<span>&#9888;</span> No figure is generated because of missing or incorrect data"
)


def figure2html(figure: BaseFigure, reactive: bool = True, close_fig: bool = False) -> str:
    r"""Convert a figure to a HTML code.

    Args:
        figure: The figure to convert.
        reactive: If ``True``, the generated is configured to be
            reactive to the screen size.
        close_fig: If ``True``, the figure is closed after it is
            converted to HTML format.

    Returns:
        The HTML code of the input figure.

    Example usage:

    ```pycon

    >>> from matplotlib import pyplot as plt
    >>> from arkas.figure import figure2html, MatplotlibFigure
    >>> fig, _ = plt.subplots()
    >>> data = figure2html(MatplotlibFigure(fig))

    ```
    """
    data = figure.set_reactive(reactive).to_html()
    if close_fig:
        figure.close()
    return data


def get_default_config() -> BaseFigureConfig:
    r"""Get the default figure config.

    Returns:
        The default figure config.

    Example usage:

    ```pycon

    >>> from arkas.figure import get_default_config
    >>> config = get_default_config()
    >>> config
    MatplotlibFigureConfig(color_norm=None)

    ```
    """
    return MatplotlibFigureConfig()



r"""Contain evaluators."""

from __future__ import annotations

__all__ = [
    "AccuracyEvaluator",
    "AveragePrecisionEvaluator",
    "BalancedAccuracyEvaluator",
    "BaseEvaluator",
    "BaseLazyEvaluator",
    "BinaryAveragePrecisionEvaluator",
    "BinaryClassificationEvaluator",
    "BinaryConfusionMatrixEvaluator",
    "BinaryFbetaScoreEvaluator",
    "BinaryJaccardEvaluator",
    "BinaryPrecisionEvaluator",
    "BinaryRecallEvaluator",
    "BinaryRocAucEvaluator",
    "EnergyDistanceEvaluator",
    "EvaluatorDict",
    "JensenShannonDivergenceEvaluator",
    "KLDivEvaluator",
    "MeanAbsoluteErrorEvaluator",
    "MeanAbsolutePercentageErrorEvaluator",
    "MeanSquaredErrorEvaluator",
    "MeanSquaredLogErrorEvaluator",
    "MeanTweedieDevianceEvaluator",
    "MedianAbsoluteErrorEvaluator",
    "MulticlassAveragePrecisionEvaluator",
    "MulticlassConfusionMatrixEvaluator",
    "MulticlassFbetaScoreEvaluator",
    "MulticlassJaccardEvaluator",
    "MulticlassPrecisionEvaluator",
    "MulticlassRecallEvaluator",
    "MulticlassRocAucEvaluator",
    "MultilabelAveragePrecisionEvaluator",
    "MultilabelConfusionMatrixEvaluator",
    "MultilabelFbetaScoreEvaluator",
    "MultilabelJaccardEvaluator",
    "MultilabelPrecisionEvaluator",
    "MultilabelRecallEvaluator",
    "MultilabelRocAucEvaluator",
    "PearsonCorrelationEvaluator",
    "R2ScoreEvaluator",
    "RootMeanSquaredErrorEvaluator",
    "SequentialEvaluator",
    "SpearmanCorrelationEvaluator",
    "WassersteinDistanceEvaluator",
    "is_evaluator_config",
    "setup_evaluator",
]

from arkas.evaluator.accuracy import AccuracyEvaluator
from arkas.evaluator.ap import AveragePrecisionEvaluator
from arkas.evaluator.balanced_accuracy import BalancedAccuracyEvaluator
from arkas.evaluator.base import BaseEvaluator, is_evaluator_config, setup_evaluator
from arkas.evaluator.binary_ap import BinaryAveragePrecisionEvaluator
from arkas.evaluator.binary_classification import BinaryClassificationEvaluator
from arkas.evaluator.binary_confmat import BinaryConfusionMatrixEvaluator
from arkas.evaluator.binary_fbeta import BinaryFbetaScoreEvaluator
from arkas.evaluator.binary_jaccard import BinaryJaccardEvaluator
from arkas.evaluator.binary_precision import BinaryPrecisionEvaluator
from arkas.evaluator.binary_recall import BinaryRecallEvaluator
from arkas.evaluator.binary_roc_auc import BinaryRocAucEvaluator
from arkas.evaluator.energy import EnergyDistanceEvaluator
from arkas.evaluator.jensen_shannon import JensenShannonDivergenceEvaluator
from arkas.evaluator.kl import KLDivEvaluator
from arkas.evaluator.lazy import BaseLazyEvaluator
from arkas.evaluator.mae import MeanAbsoluteErrorEvaluator
from arkas.evaluator.mape import MeanAbsolutePercentageErrorEvaluator
from arkas.evaluator.mapping import EvaluatorDict
from arkas.evaluator.median_error import MedianAbsoluteErrorEvaluator
from arkas.evaluator.mse import MeanSquaredErrorEvaluator
from arkas.evaluator.msle import MeanSquaredLogErrorEvaluator
from arkas.evaluator.multiclass_ap import MulticlassAveragePrecisionEvaluator
from arkas.evaluator.multiclass_confmat import MulticlassConfusionMatrixEvaluator
from arkas.evaluator.multiclass_fbeta import MulticlassFbetaScoreEvaluator
from arkas.evaluator.multiclass_jaccard import MulticlassJaccardEvaluator
from arkas.evaluator.multiclass_precision import MulticlassPrecisionEvaluator
from arkas.evaluator.multiclass_recall import MulticlassRecallEvaluator
from arkas.evaluator.multiclass_roc_auc import MulticlassRocAucEvaluator
from arkas.evaluator.multilabel_ap import MultilabelAveragePrecisionEvaluator
from arkas.evaluator.multilabel_confmat import MultilabelConfusionMatrixEvaluator
from arkas.evaluator.multilabel_fbeta import MultilabelFbetaScoreEvaluator
from arkas.evaluator.multilabel_jaccard import MultilabelJaccardEvaluator
from arkas.evaluator.multilabel_precision import MultilabelPrecisionEvaluator
from arkas.evaluator.multilabel_recall import MultilabelRecallEvaluator
from arkas.evaluator.multilabel_roc_auc import MultilabelRocAucEvaluator
from arkas.evaluator.pearson import PearsonCorrelationEvaluator
from arkas.evaluator.r2 import R2ScoreEvaluator
from arkas.evaluator.rmse import RootMeanSquaredErrorEvaluator
from arkas.evaluator.sequential import SequentialEvaluator
from arkas.evaluator.spearman import SpearmanCorrelationEvaluator
from arkas.evaluator.tweedie_deviance import MeanTweedieDevianceEvaluator
from arkas.evaluator.wasserstein import WassersteinDistanceEvaluator



r"""Contain mapping utility functions."""

from __future__ import annotations

__all__ = ["find_missing_keys"]

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence


def find_missing_keys(mapping: Mapping, keys: set | Sequence) -> set:
    r"""Return the set of keys that are not in the input mapping.

    Args:
        mapping: The input mapping.
        keys: The keys to check in the input mapping.

    Returns:
        The set of missing keys.

    Example usage:

    ```pycon

    >>> from arkas.utils.mapping import find_missing_keys
    >>> keys = find_missing_keys(
    ...     mapping={"key1": 1, "key2": 2, "key3": 3}, keys=["key1", "key2", "key4"]
    ... )
    >>> keys
    {'key4'}

    ```
    """
    keys = set(keys)
    intersection = set(mapping.keys()).intersection(keys)
    return keys.difference(intersection)



r"""Contain runners."""

from __future__ import annotations

__all__ = ["AnalysisRunner", "BaseRunner", "EvaluationRunner", "is_runner_config", "setup_runner"]

from arkas.runner.analysis import AnalysisRunner
from arkas.runner.base import BaseRunner, is_runner_config, setup_runner
from arkas.runner.evaluation import EvaluationRunner



r"""Implement the Kullback-Leibler (KL) divergence between two
distributions."""

from __future__ import annotations

__all__ = ["kl_div"]


from typing import TYPE_CHECKING

from arkas.metric.utils import preprocess_same_shape_arrays
from arkas.utils.imports import check_scipy, is_scipy_available

if is_scipy_available():
    from scipy.special import rel_entr

if TYPE_CHECKING:
    import numpy as np


def kl_div(
    p: np.ndarray,
    q: np.ndarray,
    *,
    prefix: str = "",
    suffix: str = "",
) -> dict[str, float]:
    r"""Return the Kullback-Leibler (KL) divergence between two
    distributions.

    Args:
        p: The true probability distribution.
        q: The model probability distribution.
        prefix: The key prefix in the returned dictionary.
        suffix: The key suffix in the returned dictionary.

    Returns:
        The computed metrics.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.metric import kl_div
    >>> kl_div(p=np.array([0.1, 0.6, 0.1, 0.2]), q=np.array([0.2, 0.5, 0.2, 0.1]))
    {'size': 4, 'kl_pq': 0.109..., 'kl_qp': 0.116...}

    ```
    """
    check_scipy()
    p, q = preprocess_same_shape_arrays(arrays=[p.ravel(), q.ravel()])

    size = p.size
    kl_pq, kl_qp = float("nan"), float("nan")
    if size > 0:
        kl_pq = _kl_divergence(p, q)
        kl_qp = _kl_divergence(q, p)
    return {
        f"{prefix}size{suffix}": size,
        f"{prefix}kl_pq{suffix}": kl_pq,
        f"{prefix}kl_qp{suffix}": kl_qp,
    }


def _kl_divergence(p: np.ndarray, q: np.ndarray) -> float:
    r"""Return the Kullback-Leibler (KL) divergence between two
    distributions.

    Args:
        p: The true probability distribution.
        q: The model probability distribution.

    Returns:
        The KL divergence.
    """
    return float(rel_entr(p, q).sum())



r"""Implement the precision metrics."""

from __future__ import annotations

__all__ = [
    "binary_precision",
    "find_label_type",
    "multiclass_precision",
    "multilabel_precision",
    "precision",
]

import math

import numpy as np
from sklearn import metrics

from arkas.metric.utils import (
    check_label_type,
    contains_nan,
    preprocess_pred,
    preprocess_pred_multilabel,
)


def precision(
    y_true: np.ndarray,
    y_pred: np.ndarray,
    *,
    label_type: str = "auto",
    prefix: str = "",
    suffix: str = "",
    nan_policy: str = "propagate",
) -> dict[str, float | np.ndarray]:
    r"""Return the precision metrics.

    Args:
        y_true: The ground truth target labels. This input must
            be an array of shape ``(n_samples,)`` or
            ``(n_samples, n_classes)``.
        y_pred: The predicted labels. This input must be an array of
            shape ``(n_samples,)`` or ``(n_samples, n_classes)``.
        label_type: The type of labels used to evaluate the metrics.
            The valid values are: ``'binary'``, ``'multiclass'``,
            and ``'multilabel'``. If ``'binary'`` or ``'multilabel'``,
            ``y_true`` values  must be ``0`` and ``1``.
        prefix: The key prefix in the returned dictionary.
        suffix: The key suffix in the returned dictionary.
        nan_policy: The policy on how to handle NaN values in the input
            arrays. The following options are available: ``'omit'``,
            ``'propagate'``, and ``'raise'``.

    Returns:
        The computed metrics.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.metric import precision
    >>> # auto
    >>> precision(y_true=np.array([1, 0, 0, 1, 1]), y_pred=np.array([1, 0, 0, 1, 1]))
    {'count': 5, 'precision': 1.0}
    >>> # binary
    >>> precision(
    ...     y_true=np.array([1, 0, 0, 1, 1]),
    ...     y_pred=np.array([1, 0, 0, 1, 1]),
    ...     label_type="binary",
    ... )
    {'count': 5, 'precision': 1.0}
    >>> # multiclass
    >>> precision(
    ...     y_true=np.array([0, 0, 1, 1, 2, 2]),
    ...     y_pred=np.array([0, 0, 1, 1, 2, 2]),
    ...     label_type="multiclass",
    ... )
    {'count': 6,
     'macro_precision': 1.0,
     'micro_precision': 1.0,
     'precision': array([1., 1., 1.]),
     'weighted_precision': 1.0}
    >>> # multilabel
    >>> precision(
    ...     y_true=np.array([[1, 0, 1], [0, 1, 0], [0, 1, 0], [1, 0, 1], [1, 0, 1]]),
    ...     y_pred=np.array([[1, 0, 1], [0, 1, 0], [0, 1, 0], [1, 0, 1], [1, 0, 1]]),
    ...     label_type="multilabel",
    ... )
    {'count': 5,
     'macro_precision': 1.0,
     'micro_precision': 1.0,
     'precision': array([1., 1., 1.]),
     'weighted_precision': 1.0}

    ```
    """
    check_label_type(label_type)
    if label_type == "auto":
        label_type = find_label_type(y_true=y_true, y_pred=y_pred)
    if label_type == "binary":
        return binary_precision(
            y_true=y_true, y_pred=y_pred, prefix=prefix, suffix=suffix, nan_policy=nan_policy
        )
    if label_type == "multilabel":
        return multilabel_precision(
            y_true=y_true, y_pred=y_pred, prefix=prefix, suffix=suffix, nan_policy=nan_policy
        )
    return multiclass_precision(
        y_true=y_true, y_pred=y_pred, prefix=prefix, suffix=suffix, nan_policy=nan_policy
    )


def binary_precision(
    y_true: np.ndarray,
    y_pred: np.ndarray,
    *,
    prefix: str = "",
    suffix: str = "",
    nan_policy: str = "propagate",
) -> dict[str, float]:
    r"""Return the precision metrics for binary labels.

    Args:
        y_true: The ground truth target labels. This input must
            be an array of shape ``(n_samples,)``.
        y_pred: The predicted labels. This input must
            be an array of shape ``(n_samples,)``.
        prefix: The key prefix in the returned dictionary.
        suffix: The key suffix in the returned dictionary.
        nan_policy: The policy on how to handle NaN values in the input
            arrays. The following options are available: ``'omit'``,
            ``'propagate'``, and ``'raise'``.

    Returns:
        The computed metrics.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.metric import binary_precision
    >>> binary_precision(y_true=np.array([1, 0, 0, 1, 1]), y_pred=np.array([1, 0, 0, 1, 1]))
    {'count': 5, 'precision': 1.0}

    ```
    """
    y_true, y_pred = preprocess_pred(
        y_true=y_true.ravel(), y_pred=y_pred.ravel(), drop_nan=nan_policy == "omit"
    )
    y_true_nan = contains_nan(arr=y_true, nan_policy=nan_policy, name="'y_true'")
    y_pred_nan = contains_nan(arr=y_pred, nan_policy=nan_policy, name="'y_pred'")

    count, score = y_true.size, float("nan")
    if count > 0 and not y_true_nan and not y_pred_nan:
        score = float(metrics.precision_score(y_true=y_true, y_pred=y_pred))
    return {f"{prefix}count{suffix}": count, f"{prefix}precision{suffix}": score}


def multiclass_precision(
    y_true: np.ndarray,
    y_pred: np.ndarray,
    *,
    prefix: str = "",
    suffix: str = "",
    nan_policy: str = "propagate",
) -> dict[str, float | np.ndarray]:
    r"""Return the precision metrics for multiclass labels.

    Args:
        y_true: The ground truth target labels. This input must
            be an array of shape ``(n_samples,)``.
        y_pred: The predicted labels. This input must
            be an array of shape ``(n_samples,)``.
        prefix: The key prefix in the returned dictionary.
        suffix: The key suffix in the returned dictionary.
        nan_policy: The policy on how to handle NaN values in the input
            arrays. The following options are available: ``'omit'``,
            ``'propagate'``, and ``'raise'``.

    Returns:
        The computed metrics.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.metric import multiclass_precision
    >>> multiclass_precision(
    ...     y_true=np.array([0, 0, 1, 1, 2, 2]), y_pred=np.array([0, 0, 1, 1, 2, 2])
    ... )
    {'count': 6,
     'macro_precision': 1.0,
     'micro_precision': 1.0,
     'precision': array([1., 1., 1.]),
     'weighted_precision': 1.0}

    ```
    """
    y_true, y_pred = preprocess_pred(
        y_true=y_true.ravel(), y_pred=y_pred.ravel(), drop_nan=nan_policy == "omit"
    )
    y_true_nan = contains_nan(arr=y_true, nan_policy=nan_policy, name="'y_true'")
    y_pred_nan = contains_nan(arr=y_pred, nan_policy=nan_policy, name="'y_pred'")

    per_class = np.array([])
    macro, micro, weighted = float("nan"), float("nan"), float("nan")
    n_samples = y_true.shape[0]
    if n_samples > 0 and not y_true_nan and not y_pred_nan:
        macro = metrics.precision_score(
            y_true=y_true, y_pred=y_pred, average="macro", zero_division=0.0
        )
        micro = metrics.precision_score(
            y_true=y_true, y_pred=y_pred, average="micro", zero_division=0.0
        )
        weighted = metrics.precision_score(
            y_true=y_true, y_pred=y_pred, average="weighted", zero_division=0.0
        )
        per_class = np.asarray(
            metrics.precision_score(y_true=y_true, y_pred=y_pred, average=None, zero_division=0.0)
        ).ravel()
    return {
        f"{prefix}count{suffix}": n_samples,
        f"{prefix}macro_precision{suffix}": float(macro),
        f"{prefix}micro_precision{suffix}": float(micro),
        f"{prefix}precision{suffix}": per_class,
        f"{prefix}weighted_precision{suffix}": float(weighted),
    }


def multilabel_precision(
    y_true: np.ndarray,
    y_pred: np.ndarray,
    *,
    prefix: str = "",
    suffix: str = "",
    nan_policy: str = "propagate",
) -> dict[str, float | np.ndarray]:
    r"""Return the precision metrics for multilabel labels.

    Args:
        y_true: The ground truth target labels. This input must
            be an array of shape ``(n_samples, n_classes)``.
        y_pred: The predicted labels. This input must
            be an array of shape ``(n_samples, n_classes)``.
        prefix: The key prefix in the returned dictionary.
        suffix: The key suffix in the returned dictionary.
        nan_policy: The policy on how to handle NaN values in the input
            arrays. The following options are available: ``'omit'``,
            ``'propagate'``, and ``'raise'``.

    Returns:
        The computed metrics.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.metric import multilabel_precision
    >>> multilabel_precision(
    ...     y_true=np.array([[1, 0, 1], [0, 1, 0], [0, 1, 0], [1, 0, 1], [1, 0, 1]]),
    ...     y_pred=np.array([[1, 0, 1], [0, 1, 0], [0, 1, 0], [1, 0, 1], [1, 0, 1]]),
    ... )
    {'count': 5,
     'macro_precision': 1.0,
     'micro_precision': 1.0,
     'precision': array([1., 1., 1.]),
     'weighted_precision': 1.0}

    ```
    """
    y_true, y_pred = preprocess_pred_multilabel(y_true, y_pred, drop_nan=nan_policy == "omit")
    y_true_nan = contains_nan(arr=y_true, nan_policy=nan_policy, name="'y_true'")
    y_pred_nan = contains_nan(arr=y_pred, nan_policy=nan_policy, name="'y_pred'")

    per_class = np.array([])
    macro, micro, weighted = float("nan"), float("nan"), float("nan")
    n_samples = y_true.shape[0]
    if n_samples > 0 and not y_true_nan and not y_pred_nan:
        per_class = np.array(
            metrics.precision_score(
                y_true=y_true,
                y_pred=y_pred,
                average="binary" if y_pred.shape[1] == 1 else None,
            )
        ).ravel()
        macro = metrics.precision_score(y_true=y_true, y_pred=y_pred, average="macro")
        micro = metrics.precision_score(y_true=y_true, y_pred=y_pred, average="micro")
        weighted = metrics.precision_score(y_true=y_true, y_pred=y_pred, average="weighted")

    return {
        f"{prefix}count{suffix}": n_samples,
        f"{prefix}macro_precision{suffix}": float(macro),
        f"{prefix}micro_precision{suffix}": float(micro),
        f"{prefix}precision{suffix}": per_class,
        f"{prefix}weighted_precision{suffix}": float(weighted),
    }


def find_label_type(y_true: np.ndarray, y_pred: np.ndarray) -> str:
    r"""Try to find the label type automatically based on the arrays'
    shape and values.

    Note:
        NaN are used to indicate invalid/missing values.

    Args:
        y_true: The ground truth target labels.
        y_pred: The predicted labels.

    Returns:
        The label type.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.metric.classification.precision import find_label_type
    >>> # binary
    >>> label_type = find_label_type(
    ...     y_true=np.array([1, 0, 0, 1, 1]), y_pred=np.array([1, 0, 0, 1, 1])
    ... )
    >>> label_type
    'binary'
    >>> # multiclass
    >>> label_type = find_label_type(
    ...     y_true=np.array([0, 0, 1, 1, 2, 2]), y_pred=np.array([0, 0, 1, 1, 2, 2])
    ... )
    >>> label_type
    'multiclass'

    ```
    """
    # remove NaNs because they indicate missing values
    unique = set(filter(lambda x: not math.isnan(x), np.unique(y_true).tolist()))
    if unique.issubset({0, 1}):
        if y_true.ndim == 2 and y_pred.ndim == 2:
            return "multilabel"
        return "binary"
    return "multiclass"



r"""Implement the absolute error metrics."""

from __future__ import annotations

__all__ = ["mean_absolute_error", "median_absolute_error"]


from typing import TYPE_CHECKING

from sklearn import metrics

from arkas.metric.utils import contains_nan, preprocess_pred

if TYPE_CHECKING:
    import numpy as np


def mean_absolute_error(
    y_true: np.ndarray,
    y_pred: np.ndarray,
    *,
    prefix: str = "",
    suffix: str = "",
    nan_policy: str = "propagate",
) -> dict[str, float]:
    r"""Return the mean absolute error (MAE).

    Args:
        y_true: The ground truth target values.
        y_pred: The predicted values.
        prefix: The key prefix in the returned dictionary.
        suffix: The key suffix in the returned dictionary.
        nan_policy: The policy on how to handle NaN values in the input
            arrays. The following options are available: ``'omit'``,
            ``'propagate'``, and ``'raise'``.

    Returns:
        The computed metrics.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.metric import mean_absolute_error
    >>> mean_absolute_error(y_true=np.array([1, 2, 3, 4, 5]), y_pred=np.array([1, 2, 3, 4, 5]))
    {'count': 5, 'mean_absolute_error': 0.0}

    ```
    """
    y_true, y_pred = preprocess_pred(
        y_true=y_true.ravel(), y_pred=y_pred.ravel(), drop_nan=nan_policy == "omit"
    )
    y_true_nan = contains_nan(arr=y_true, nan_policy=nan_policy, name="'y_true'")
    y_pred_nan = contains_nan(arr=y_pred, nan_policy=nan_policy, name="'y_pred'")

    count = y_true.size
    error = float("nan")
    if count > 0 and not y_true_nan and not y_pred_nan:
        error = float(metrics.mean_absolute_error(y_true=y_true, y_pred=y_pred))
    return {
        f"{prefix}count{suffix}": count,
        f"{prefix}mean_absolute_error{suffix}": error,
    }


def median_absolute_error(
    y_true: np.ndarray,
    y_pred: np.ndarray,
    *,
    prefix: str = "",
    suffix: str = "",
    nan_policy: str = "propagate",
) -> dict[str, float]:
    r"""Return the median absolute error.

    Args:
        y_true: The ground truth target values.
        y_pred: The predicted values.
        prefix: The key prefix in the returned dictionary.
        suffix: The key suffix in the returned dictionary.
        nan_policy: The policy on how to handle NaN values in the input
            arrays. The following options are available: ``'omit'``,
            ``'propagate'``, and ``'raise'``.

    Returns:
        The computed metrics.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.metric import median_absolute_error
    >>> median_absolute_error(
    ...     y_true=np.array([1, 2, 3, 4, 5]), y_pred=np.array([1, 2, 3, 4, 5])
    ... )
    {'count': 5, 'median_absolute_error': 0.0}

    ```
    """
    y_true, y_pred = preprocess_pred(
        y_true=y_true.ravel(), y_pred=y_pred.ravel(), drop_nan=nan_policy == "omit"
    )
    y_true_nan = contains_nan(arr=y_true, nan_policy=nan_policy, name="'y_true'")
    y_pred_nan = contains_nan(arr=y_pred, nan_policy=nan_policy, name="'y_pred'")

    count = y_true.size
    error = float("nan")
    if count > 0 and not y_true_nan and not y_pred_nan:
        error = float(metrics.median_absolute_error(y_true=y_true, y_pred=y_pred))
    return {
        f"{prefix}count{suffix}": count,
        f"{prefix}median_absolute_error{suffix}": error,
    }



r"""Implement some utility functions to manage optional dependencies."""

from __future__ import annotations

__all__ = [
    "check_colorlog",
    "check_hya",
    "check_hydra",
    "check_markdown",
    "check_matplotlib",
    "check_omegaconf",
    "check_plotly",
    "check_scipy",
    "colorlog_available",
    "hya_available",
    "hydra_available",
    "is_colorlog_available",
    "is_hya_available",
    "is_hydra_available",
    "is_markdown_available",
    "is_matplotlib_available",
    "is_omegaconf_available",
    "is_plotly_available",
    "is_scipy_available",
    "markdown_available",
    "matplotlib_available",
    "omegaconf_available",
    "plotly_available",
    "scipy_available",
]

from typing import TYPE_CHECKING, Any

from coola.utils.imports import decorator_package_available, package_available

if TYPE_CHECKING:
    from collections.abc import Callable


####################
#     colorlog     #
####################


def is_colorlog_available() -> bool:
    r"""Indicate if the ``colorlog`` package is installed or not.

    Returns:
        ``True`` if ``colorlog`` is available otherwise
            ``False``.

    Example usage:

    ```pycon

    >>> from arkas.utils.imports import is_colorlog_available
    >>> is_colorlog_available()

    ```
    """
    return package_available("colorlog")


def check_colorlog() -> None:
    r"""Check if the ``colorlog`` package is installed.

    Raises:
        RuntimeError: if the ``colorlog`` package is not
            installed.

    Example usage:

    ```pycon

    >>> from arkas.utils.imports import check_colorlog
    >>> check_colorlog()

    ```
    """
    if not is_colorlog_available():
        msg = (
            "'colorlog' package is required but not installed. "
            "You can install 'colorlog' package with the command:\n\n"
            "pip install colorlog\n"
        )
        raise RuntimeError(msg)


def colorlog_available(fn: Callable[..., Any]) -> Callable[..., Any]:
    r"""Implement a decorator to execute a function only if ``colorlog``
    package is installed.

    Args:
        fn: The function to execute.

    Returns:
        A wrapper around ``fn`` if ``colorlog`` package is
            installed, otherwise ``None``.

    Example usage:

    ```pycon

    >>> from arkas.utils.imports import colorlog_available
    >>> @colorlog_available
    ... def my_function(n: int = 0) -> int:
    ...     return 42 + n
    ...
    >>> my_function()

    ```
    """
    return decorator_package_available(fn, is_colorlog_available)


###############
#     hya     #
###############


def is_hya_available() -> bool:
    r"""Indicate if the ``hya`` package is installed or not.

    Returns:
        ``True`` if ``hya`` is available otherwise
            ``False``.

    Example usage:

    ```pycon

    >>> from arkas.utils.imports import is_hya_available
    >>> is_hya_available()

    ```
    """
    return package_available("hya")


def check_hya() -> None:
    r"""Check if the ``hya`` package is installed.

    Raises:
        RuntimeError: if the ``hya`` package is not
            installed.

    Example usage:

    ```pycon

    >>> from arkas.utils.imports import check_hya
    >>> check_hya()

    ```
    """
    if not is_hya_available():
        msg = (
            "'hya' package is required but not installed. "
            "You can install 'hya' package with the command:\n\n"
            "pip install hya\n"
        )
        raise RuntimeError(msg)


def hya_available(fn: Callable[..., Any]) -> Callable[..., Any]:
    r"""Implement a decorator to execute a function only if ``hya``
    package is installed.

    Args:
        fn: The function to execute.

    Returns:
        A wrapper around ``fn`` if ``hya`` package is
            installed, otherwise ``None``.

    Example usage:

    ```pycon

    >>> from arkas.utils.imports import hya_available
    >>> @hya_available
    ... def my_function(n: int = 0) -> int:
    ...     return 42 + n
    ...
    >>> my_function()

    ```
    """
    return decorator_package_available(fn, is_hya_available)


#################
#     hydra     #
#################


def is_hydra_available() -> bool:
    r"""Indicate if the ``hydra`` package is installed or not.

    Returns:
        ``True`` if ``hydra`` is available otherwise
            ``False``.

    Example usage:

    ```pycon

    >>> from arkas.utils.imports import is_hydra_available
    >>> is_hydra_available()

    ```
    """
    return package_available("hydra")


def check_hydra() -> None:
    r"""Check if the ``hydra`` package is installed.

    Raises:
        RuntimeError: if the ``hydra`` package is not
            installed.

    Example usage:

    ```pycon

    >>> from arkas.utils.imports import check_hydra
    >>> check_hydra()

    ```
    """
    if not is_hydra_available():
        msg = (
            "'hydra' package is required but not installed. "
            "You can install 'hydra' package with the command:\n\n"
            "pip install hydra-core\n"
        )
        raise RuntimeError(msg)


def hydra_available(fn: Callable[..., Any]) -> Callable[..., Any]:
    r"""Implement a decorator to execute a function only if ``hydra``
    package is installed.

    Args:
        fn: The function to execute.

    Returns:
        A wrapper around ``fn`` if ``hydra`` package is
            installed, otherwise ``None``.

    Example usage:

    ```pycon

    >>> from arkas.utils.imports import hydra_available
    >>> @hydra_available
    ... def my_function(n: int = 0) -> int:
    ...     return 42 + n
    ...
    >>> my_function()

    ```
    """
    return decorator_package_available(fn, is_hydra_available)


####################
#     markdown     #
####################


def is_markdown_available() -> bool:
    r"""Indicate if the ``markdown`` package is installed or not.

    Returns:
        ``True`` if ``markdown`` is available otherwise
            ``False``.

    Example usage:

    ```pycon

    >>> from arkas.utils.imports import is_markdown_available
    >>> is_markdown_available()

    ```
    """
    return package_available("markdown")


def check_markdown() -> None:
    r"""Check if the ``markdown`` package is installed.

    Raises:
        RuntimeError: if the ``markdown`` package is not
            installed.

    Example usage:

    ```pycon

    >>> from arkas.utils.imports import check_markdown
    >>> check_markdown()

    ```
    """
    if not is_markdown_available():
        msg = (
            "'markdown' package is required but not installed. "
            "You can install 'markdown' package with the command:\n\n"
            "pip install markdown\n"
        )
        raise RuntimeError(msg)


def markdown_available(fn: Callable[..., Any]) -> Callable[..., Any]:
    r"""Implement a decorator to execute a function only if ``markdown``
    package is installed.

    Args:
        fn: The function to execute.

    Returns:
        A wrapper around ``fn`` if ``markdown`` package is
            installed, otherwise ``None``.

    Example usage:

    ```pycon

    >>> from arkas.utils.imports import markdown_available
    >>> @markdown_available
    ... def my_function(n: int = 0) -> int:
    ...     return 42 + n
    ...
    >>> my_function()

    ```
    """
    return decorator_package_available(fn, is_markdown_available)


######################
#     matplotlib     #
######################


def is_matplotlib_available() -> bool:
    r"""Indicate if the ``matplotlib`` package is installed or not.

    Returns:
        ``True`` if ``matplotlib`` is available otherwise
            ``False``.

    Example usage:

    ```pycon

    >>> from arkas.utils.imports import is_matplotlib_available
    >>> is_matplotlib_available()

    ```
    """
    return package_available("matplotlib")


def check_matplotlib() -> None:
    r"""Check if the ``matplotlib`` package is installed.

    Raises:
        RuntimeError: if the ``matplotlib`` package is not
            installed.

    Example usage:

    ```pycon

    >>> from arkas.utils.imports import check_matplotlib
    >>> check_matplotlib()

    ```
    """
    if not is_matplotlib_available():
        msg = (
            "'matplotlib' package is required but not installed. "
            "You can install 'matplotlib' package with the command:\n\n"
            "pip install matplotlib\n"
        )
        raise RuntimeError(msg)


def matplotlib_available(fn: Callable[..., Any]) -> Callable[..., Any]:
    r"""Implement a decorator to execute a function only if
    ``matplotlib`` package is installed.

    Args:
        fn: The function to execute.

    Returns:
        A wrapper around ``fn`` if ``matplotlib`` package is
            installed, otherwise ``None``.

    Example usage:

    ```pycon

    >>> from arkas.utils.imports import matplotlib_available
    >>> @matplotlib_available
    ... def my_function(n: int = 0) -> int:
    ...     return 42 + n
    ...
    >>> my_function()

    ```
    """
    return decorator_package_available(fn, is_matplotlib_available)


#####################
#     omegaconf     #
#####################


def is_omegaconf_available() -> bool:
    r"""Indicate if the ``omegaconf`` package is installed or not.

    Returns:
        ``True`` if ``omegaconf`` is available otherwise
            ``False``.

    Example usage:

    ```pycon

    >>> from arkas.utils.imports import is_omegaconf_available
    >>> is_omegaconf_available()

    ```
    """
    return package_available("omegaconf")


def check_omegaconf() -> None:
    r"""Check if the ``omegaconf`` package is installed.

    Raises:
        RuntimeError: if the ``omegaconf`` package is not
            installed.

    Example usage:

    ```pycon

    >>> from arkas.utils.imports import check_omegaconf
    >>> check_omegaconf()

    ```
    """
    if not is_omegaconf_available():
        msg = (
            "'omegaconf' package is required but not installed. "
            "You can install 'omegaconf' package with the command:\n\n"
            "pip install omegaconf\n"
        )
        raise RuntimeError(msg)


def omegaconf_available(fn: Callable[..., Any]) -> Callable[..., Any]:
    r"""Implement a decorator to execute a function only if ``omegaconf``
    package is installed.

    Args:
        fn: The function to execute.

    Returns:
        A wrapper around ``fn`` if ``omegaconf`` package is
            installed, otherwise ``None``.

    Example usage:

    ```pycon

    >>> from arkas.utils.imports import omegaconf_available
    >>> @omegaconf_available
    ... def my_function(n: int = 0) -> int:
    ...     return 42 + n
    ...
    >>> my_function()

    ```
    """
    return decorator_package_available(fn, is_omegaconf_available)


##################
#     plotly     #
##################


def is_plotly_available() -> bool:
    r"""Indicate if the ``plotly`` package is installed or not.

    Returns:
        ``True`` if ``plotly`` is available otherwise
            ``False``.

    Example usage:

    ```pycon

    >>> from arkas.utils.imports import is_plotly_available
    >>> is_plotly_available()

    ```
    """
    return package_available("plotly")


def check_plotly() -> None:
    r"""Check if the ``plotly`` package is installed.

    Raises:
        RuntimeError: if the ``plotly`` package is not
            installed.

    Example usage:

    ```pycon

    >>> from arkas.utils.imports import check_plotly
    >>> check_plotly()

    ```
    """
    if not is_plotly_available():
        msg = (
            "'plotly' package is required but not installed. "
            "You can install 'plotly' package with the command:\n\n"
            "pip install plotly\n"
        )
        raise RuntimeError(msg)


def plotly_available(fn: Callable[..., Any]) -> Callable[..., Any]:
    r"""Implement a decorator to execute a function only if ``plotly``
    package is installed.

    Args:
        fn: The function to execute.

    Returns:
        A wrapper around ``fn`` if ``plotly`` package is
            installed, otherwise ``None``.

    Example usage:

    ```pycon

    >>> from arkas.utils.imports import plotly_available
    >>> @plotly_available
    ... def my_function(n: int = 0) -> int:
    ...     return 42 + n
    ...
    >>> my_function()

    ```
    """
    return decorator_package_available(fn, is_plotly_available)


#################
#     scipy     #
#################


def is_scipy_available() -> bool:
    r"""Indicate if the ``scipy`` package is installed or not.

    Returns:
        ``True`` if ``scipy`` is available otherwise
            ``False``.

    Example usage:

    ```pycon

    >>> from arkas.utils.imports import is_scipy_available
    >>> is_scipy_available()

    ```
    """
    return package_available("scipy")


def check_scipy() -> None:
    r"""Check if the ``scipy`` package is installed.

    Raises:
        RuntimeError: if the ``scipy`` package is not
            installed.

    Example usage:

    ```pycon

    >>> from arkas.utils.imports import check_scipy
    >>> check_scipy()

    ```
    """
    if not is_scipy_available():
        msg = (
            "'scipy' package is required but not installed. "
            "You can install 'scipy' package with the command:\n\n"
            "pip install scipy\n"
        )
        raise RuntimeError(msg)


def scipy_available(fn: Callable[..., Any]) -> Callable[..., Any]:
    r"""Implement a decorator to execute a function only if ``scipy``
    package is installed.

    Args:
        fn: The function to execute.

    Returns:
        A wrapper around ``fn`` if ``scipy`` package is
            installed, otherwise ``None``.

    Example usage:

    ```pycon

    >>> from arkas.utils.imports import scipy_available
    >>> @scipy_available
    ... def my_function(n: int = 0) -> int:
    ...     return 42 + n
    ...
    >>> my_function()

    ```
    """
    return decorator_package_available(fn, is_scipy_available)



r"""Contain the base class to represent a run."""

from __future__ import annotations

__all__ = ["BaseRun"]

from abc import abstractmethod
from typing import Any


class BaseRun:
    r"""Define the base class to represent a run."""

    @abstractmethod
    def get_uri(self) -> str:
        r"""Get the Uniform Resource Identifier (URI) of the run.

        Returns:
            The Uniform Resource Identifier (URI).
        """

    @abstractmethod
    def get_data(self, name: str) -> Any:
        r"""Get data.

        Args:
            name: The name of the data to get.

        Returns:
            The data associated to the name.
        """

    @abstractmethod
    def get_metrics(self) -> dict:
        r"""Get the metric values.

        Returns:
            The metric values.
        """



r"""Contain utility functions to generate plots."""

from __future__ import annotations

__all__ = [
    "auto_yscale_continuous",
    "auto_yscale_discrete",
    "axvline_median",
    "axvline_quantile",
    "readable_xticklabels",
    "readable_yticklabels",
]

from arkas.plot.utils.line import axvline_median, axvline_quantile
from arkas.plot.utils.scale import auto_yscale_continuous, auto_yscale_discrete
from arkas.plot.utils.tick import readable_xticklabels, readable_yticklabels



r"""Contain the implementation of a pairwise column co-occurrence
plotter."""

from __future__ import annotations

__all__ = ["BaseFigureCreator", "ColumnCooccurrencePlotter", "MatplotlibFigureCreator"]

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any

import matplotlib.pyplot as plt
from coola import objects_are_equal
from grizz.utils.cooccurrence import compute_pairwise_cooccurrence

from arkas.figure.creator import FigureCreatorRegistry
from arkas.figure.html import HtmlFigure
from arkas.figure.matplotlib import MatplotlibFigure, MatplotlibFigureConfig
from arkas.figure.utils import MISSING_FIGURE_MESSAGE, get_default_config
from arkas.plot.utils import readable_xticklabels, readable_yticklabels
from arkas.plotter.base import BasePlotter
from arkas.plotter.vanilla import Plotter

if TYPE_CHECKING:
    from collections.abc import Sequence

    import numpy as np
    import polars as pl

    from arkas.figure.base import BaseFigure, BaseFigureConfig


class BaseFigureCreator(ABC):
    r"""Define the base class to create a figure of the pairwise column
    co-occurrence matrix.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.figure import MatplotlibFigureConfig
    >>> from arkas.plotter.column_cooccurrence import MatplotlibFigureCreator
    >>> creator = MatplotlibFigureCreator()
    >>> creator
    MatplotlibFigureCreator()
    >>> config = MatplotlibFigureConfig()
    >>> fig = creator.create(matrix=np.ones((3, 3)), columns=["a", "b", "c"], config=config)

    ```
    """

    @abstractmethod
    def create(
        self, matrix: np.ndarray, columns: Sequence[str], config: BaseFigureConfig
    ) -> BaseFigure:
        r"""Create a figure of the pairwise column co-occurrence matrix.

        Args:
            matrix: The co-occurrence matrix.
            columns: The column names.
            config: The figure config.

        Returns:
            The generated figure.

        Example usage:

        ```pycon

        >>> import numpy as np
        >>> from arkas.figure import MatplotlibFigureConfig
        >>> from arkas.plotter.column_cooccurrence import MatplotlibFigureCreator
        >>> creator = MatplotlibFigureCreator()
        >>> config = MatplotlibFigureConfig()
        >>> fig = creator.create(matrix=np.ones((3, 3)), columns=["a", "b", "c"], config=config)

        ```
        """


class MatplotlibFigureCreator(BaseFigureCreator):
    r"""Create a matplotlib figure of the pairwise column co-occurrence
    matrix.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.figure import MatplotlibFigureConfig
    >>> from arkas.plotter.column_cooccurrence import MatplotlibFigureCreator
    >>> creator = MatplotlibFigureCreator()
    >>> creator
    MatplotlibFigureCreator()
    >>> config = MatplotlibFigureConfig()
    >>> fig = creator.create(matrix=np.ones((3, 3)), columns=["a", "b", "c"], config=config)

    ```
    """

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}()"

    def create(
        self, matrix: np.ndarray, columns: Sequence[str], config: MatplotlibFigureConfig
    ) -> BaseFigure:
        if matrix.shape[0] == 0:
            return HtmlFigure(MISSING_FIGURE_MESSAGE)

        fig, ax = plt.subplots(**config.get_args())
        im = ax.imshow(matrix, norm=config.get_color_norm())
        fig.colorbar(im)
        ax.set_xticks(
            range(len(columns)),
            labels=columns,
            rotation=45,
            ha="right",
            rotation_mode="anchor",
        )
        ax.set_yticks(
            range(len(columns)),
            labels=columns,
        )
        readable_xticklabels(ax, max_num_xticks=50)
        readable_yticklabels(ax, max_num_yticks=50)
        ax.set_title("pairwise column co-occurrence matrix")

        if matrix.shape[0] < 16:
            for i in range(len(columns)):
                for j in range(len(columns)):
                    ax.text(
                        j, i, matrix[i, j], ha="center", va="center", color="w", fontsize="xx-small"
                    )

        fig.tight_layout()
        return MatplotlibFigure(fig)


class ColumnCooccurrencePlotter(BasePlotter):
    r"""Implement a pairwise column co-occurrence plotter.

    Args:
        frame: The DataFrame to analyze.
        ignore_self: If ``True``, the diagonal of the co-occurrence
            matrix (a.k.a. self-co-occurrence) is set to 0.
        figure_config: The figure configuration.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from arkas.plotter import ColumnCooccurrencePlotter
    >>> frame = pl.DataFrame(
    ...     {
    ...         "col1": [0, 1, 1, 0, 0, 1, 0],
    ...         "col2": [0, 1, 0, 1, 0, 1, 0],
    ...         "col3": [0, 0, 0, 0, 1, 1, 1],
    ...     }
    ... )
    >>> plotter = ColumnCooccurrencePlotter(frame)
    >>> plotter
    ColumnCooccurrencePlotter(shape=(7, 3), ignore_self=False)

    ```
    """

    registry: FigureCreatorRegistry = FigureCreatorRegistry(
        {MatplotlibFigureConfig.backend(): MatplotlibFigureCreator()}
    )

    def __init__(
        self,
        frame: pl.DataFrame,
        ignore_self: bool = False,
        figure_config: BaseFigureConfig | None = None,
    ) -> None:
        self._frame = frame
        self._ignore_self = bool(ignore_self)
        self._figure_config = figure_config or get_default_config()

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(shape={self._frame.shape}, "
            f"ignore_self={self._ignore_self})"
        )

    def compute(self) -> Plotter:
        return Plotter(self.plot())

    def equal(self, other: Any, equal_nan: bool = False) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return (
            self._ignore_self == other._ignore_self
            and objects_are_equal(self._frame, other._frame, equal_nan=equal_nan)
            and self._figure_config.equal(other._figure_config)
        )

    def plot(self, prefix: str = "", suffix: str = "") -> dict:
        figure = self.registry.find_creator(self._figure_config.backend()).create(
            matrix=self.cooccurrence_matrix(),
            columns=self._frame.columns,
            config=self._figure_config,
        )
        return {f"{prefix}column_cooccurrence{suffix}": figure}

    def cooccurrence_matrix(self) -> np.ndarray:
        r"""Return the pairwise column co-occurrence matrix.

        Returns:
            The pairwise column co-occurrence.
        """
        return compute_pairwise_cooccurrence(frame=self._frame, ignore_self=self._ignore_self)



r"""Contain the base class to implement a section."""

from __future__ import annotations

__all__ = ["BaseSection"]

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from collections.abc import Sequence


class BaseSection(ABC):
    r"""Define the base class to manage sections."""

    @abstractmethod
    def equal(self, other: Any) -> bool:
        r"""Indicate if two sections are equal or not.

        Args:
            other: The other section to compare.

        Returns:
            ``True`` if the two sections are equal, otherwise ``False``.
        """

    @abstractmethod
    def generate_html_body(self, number: str = "", tags: Sequence[str] = (), depth: int = 0) -> str:
        r"""Return the HTML body associated to the section.

        Args:
            number: The section number.
            tags: The tags associated to the section.
            depth: The depth in the report.

        Returns:
            The HTML body associated to the section.
        """

    @abstractmethod
    def generate_html_toc(
        self, number: str = "", tags: Sequence[str] = (), depth: int = 0, max_depth: int = 1
    ) -> str:
        r"""Return the HTML table of content (TOC) associated to the
        section.

        Args:
            number: The section number associated to the
                section.
            tags: The tags associated to the section.
            depth: The depth in the report.
            max_depth: The maximum depth to generate in the TOC.

        Returns:
            The HTML table of content associated to the section.
        """



r"""Contain the implementation of a HTML content generator that returns
a summary of a DataFrame."""

from __future__ import annotations

__all__ = [
    "DataFrameSummaryContentGenerator",
    "create_table",
    "create_table_row",
    "create_template",
]

import logging
from collections import Counter
from typing import TYPE_CHECKING, Any

from coola import objects_are_equal
from grizz.utils.count import compute_nunique
from grizz.utils.null import compute_null_count
from jinja2 import Template

from arkas.content.section import BaseSectionContentGenerator
from arkas.utils.validation import check_positive

if TYPE_CHECKING:
    from collections.abc import Sequence

    import polars as pl

logger = logging.getLogger(__name__)


class DataFrameSummaryContentGenerator(BaseSectionContentGenerator):
    r"""Implement a content generator that returns a summary of a
    DataFrame.

    Args:
        frame: The DataFrame to analyze.
        top: The number of most frequent values to show.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from arkas.content import DataFrameSummaryContentGenerator
    >>> content = DataFrameSummaryContentGenerator(
    ...     frame=pl.DataFrame(
    ...         {
    ...             "col1": [1.2, 4.2, 4.2, 2.2],
    ...             "col2": [1, 1, 1, 1],
    ...             "col3": [1, 2, 2, 2],
    ...         },
    ...         schema={"col1": pl.Float64, "col2": pl.Int64, "col3": pl.Int64},
    ...     )
    ... )
    >>> content
    DataFrameSummaryContentGenerator(shape=(4, 3), top=5)

    ```
    """

    def __init__(self, frame: pl.DataFrame, top: int = 5) -> None:
        self._frame = frame
        check_positive(name="top", value=top)
        self._top = top

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(shape={self._frame.shape}, top={self._top})"

    @property
    def frame(self) -> pl.DataFrame:
        r"""The DataFrame to analyze."""
        return self._frame

    @property
    def top(self) -> int:
        return self._top

    def equal(self, other: Any, equal_nan: bool = False) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return self.top == other.top and objects_are_equal(
            self.frame, other.frame, equal_nan=equal_nan
        )

    def get_columns(self) -> tuple[str, ...]:
        return tuple(self._frame.columns)

    def get_null_count(self) -> tuple[int, ...]:
        return tuple(compute_null_count(self._frame).tolist())

    def get_nunique(self) -> tuple[int, ...]:
        return tuple(compute_nunique(self._frame).tolist())

    def get_dtypes(self) -> tuple[pl.DataType, ...]:
        return tuple(self._frame.schema.dtypes())

    def get_most_frequent_values(self, top: int = 5) -> tuple[tuple[tuple[Any, int], ...], ...]:
        return tuple(tuple(Counter(series.to_list()).most_common(top)) for series in self.frame)

    def generate_content(self) -> str:
        logger.info("Generating the DataFrame summary content...")
        return Template(create_template()).render(
            {
                "table": self._create_table(),
                "nrows": f"{self._frame.shape[0]:,}",
                "ncols": f"{self._frame.shape[1]:,}",
            }
        )

    def _create_table(self) -> str:
        return create_table(
            columns=self.get_columns(),
            null_count=self.get_null_count(),
            nunique=self.get_nunique(),
            dtypes=self.get_dtypes(),
            most_frequent_values=self.get_most_frequent_values(top=self._top),
            total=self._frame.shape[0],
        )


def create_template() -> str:
    r"""Return the template of the content.

    Returns:
        The content template.

    Example usage:

    ```pycon

    >>> from arkas.content.frame_summary import create_template
    >>> template = create_template()

    ```
    """
    return """This section shows a short summary of each column.

<ul>
  <li> <b>column</b>: are the column names</li>
  <li> <b>types</b>: are the object types for the objects in the column </li>
  <li> <b>null</b>: are the number (and percentage) of null values in the column </li>
  <li> <b>unique</b>: are the number (and percentage) of unique values in the column </li>
</ul>

<p style="margin-top: 1rem;">
<b>General statistics about the DataFrame</b>

<ul>
  <li> number of columns: {{ncols}} </li>
  <li> number of rows: {{nrows}}</li>
</ul>

{{table}}
"""


def create_table(
    columns: Sequence[str],
    null_count: Sequence[int],
    nunique: Sequence[int],
    dtypes: Sequence[pl.DataType],
    most_frequent_values: Sequence[Sequence[tuple[Any, int]]],
    total: int,
) -> str:
    r"""Return a HTML representation of a table with the temporal
    distribution of null values.

    Args:
        columns: The column names.
        null_count: The number of null values for each column.
        nunique: The number of unique values for each column.
        dtypes: The data type for each column.
        most_frequent_values: The most frequent values for each column.
        total: The total number of rows.

    Returns:
        The HTML representation of the table.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from arkas.content.frame_summary import create_table_row
    >>> row = create_table(
    ...     columns=["float", "int", "str"],
    ...     null_count=(1, 0, 2),
    ...     nunique=(5, 2, 4),
    ...     dtypes=(pl.Float64(), pl.Int64(), pl.String()),
    ...     most_frequent_values=(
    ...         ((2.2, 2), (1.2, 1), (4.2, 1), (None, 1), (1.0, 1)),
    ...         ((1, 5), (0, 1)),
    ...         (("B", 2), (None, 2), ("A", 1), ("C", 1)),
    ...     ),
    ...     total=42,
    ... )

    ```
    """
    rows = []
    for (
        column,
        null,
        nuniq,
        dtype,
        mf_values,
    ) in zip(columns, null_count, nunique, dtypes, most_frequent_values):
        rows.append(
            create_table_row(
                column=column,
                null=null,
                dtype=dtype,
                nunique=nuniq,
                most_frequent_values=mf_values,
                total=total,
            )
        )
    rows = "\n".join(rows)
    return Template(
        """<table class="table table-hover table-responsive w-auto" >
    <thead class="thead table-group-divider">
        <tr>
            <th>column</th>
            <th>types</th>
            <th>null</th>
            <th>unique</th>
            <th>most frequent values</th>
        </tr>
    </thead>
    <tbody class="tbody table-group-divider">
        {{rows}}
        <tr class="table-group-divider"></tr>
    </tbody>
</table>
"""
    ).render({"rows": rows})


def create_table_row(
    column: str,
    null: int,
    nunique: int,
    dtype: pl.DataType,
    most_frequent_values: Sequence[tuple[Any, int]],
    total: int,
) -> str:
    r"""Create the HTML code of a new table row.

    Args:
        column: The column name.
        null: The number of null values.
        nunique: The number of unique values.
        dtype: The data type of the column.
        most_frequent_values: The most frequent values.
        total: The total number of rows.

    Returns:
        The HTML code of a row.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from arkas.content.frame_summary import create_table_row
    >>> row = create_table_row(
    ...     column="col",
    ...     null=5,
    ...     nunique=42,
    ...     dtype=pl.Float64(),
    ...     most_frequent_values=[("C", 12), ("A", 5), ("B", 4)],
    ...     total=100,
    ... )

    ```
    """
    null = f"{null:,} ({100 * null / total if total else float('nan'):.2f}%)"
    nunique = f"{nunique:,} ({100 * nunique / total if total else float('nan'):.2f}%)"
    most_frequent_values = ", ".join(
        [f"{val} ({100 * c / total:.2f}%)" for val, c in most_frequent_values]
    )
    return Template(
        """<tr>
    <th>{{column}}</th>
    <td>{{dtype}}</td>
    <td {{num_style}}>{{null}}</td>
    <td {{num_style}}>{{nunique}}</td>
    <td>{{most_frequent_values}}</td>
</tr>"""
    ).render(
        {
            "num_style": 'style="text-align: right;"',
            "column": column,
            "null": null,
            "dtype": dtype,
            "nunique": nunique,
            "most_frequent_values": most_frequent_values,
        }
    )



r"""Contain a simple implementation to represent a run."""

from __future__ import annotations

__all__ = ["Run"]

from typing import Any

from arkas.run.base import BaseRun


class Run(BaseRun):
    r"""Implement a simple run representation.

    Args:
        uri: The Uniform Resource Identifier (URI) of the run.
        metrics: The metrics associated to the run.
        data: The data associated to the run.
    """

    def __init__(self, uri: str, metrics: dict | None = None, data: dict | None = None) -> None:
        self._uri = uri
        self._metrics = metrics or {}
        self._data = data or {}

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(uri={self.get_uri()})"

    def get_uri(self) -> str:
        return self._uri

    def get_data(self, name: str) -> Any:
        return self._data[name]

    def get_metrics(self) -> dict:
        return self._metrics



r"""Contain sections."""

from __future__ import annotations

__all__ = [
    "AccuracySection",
    "BaseSection",
    "BinaryPrecisionSection",
    "ContentSection",
    "ResultSection",
]

from arkas.section.accuracy import AccuracySection
from arkas.section.base import BaseSection
from arkas.section.binary_precision import BinaryPrecisionSection
from arkas.section.content import ContentSection
from arkas.section.result import ResultSection



r"""Implement the F-beta metrics."""

from __future__ import annotations

__all__ = [
    "binary_fbeta_score",
    "fbeta_score",
    "multiclass_fbeta_score",
    "multilabel_fbeta_score",
]

from typing import TYPE_CHECKING

import numpy as np
from sklearn import metrics

from arkas.metric.classification.precision import find_label_type
from arkas.metric.utils import (
    check_label_type,
    contains_nan,
    preprocess_pred,
    preprocess_pred_multilabel,
)

if TYPE_CHECKING:
    from collections.abc import Sequence


def fbeta_score(
    y_true: np.ndarray,
    y_pred: np.ndarray,
    *,
    betas: Sequence[float] = (1,),
    label_type: str = "auto",
    prefix: str = "",
    suffix: str = "",
    nan_policy: str = "propagate",
) -> dict[str, float | np.ndarray]:
    r"""Return the F-beta metrics.

    Args:
        y_true: The ground truth target labels. This input must
            be an array of shape ``(n_samples,)`` or
            ``(n_samples, n_classes)``.
        y_pred: The predicted labels. This input must be an array of
            shape ``(n_samples,)`` or ``(n_samples, n_classes)``.
        betas: The betas used to compute the F-beta scores.
        label_type: The type of labels used to evaluate the metrics.
            The valid values are: ``'binary'``, ``'multiclass'``,
            and ``'multilabel'``. If ``'binary'`` or ``'multilabel'``,
            ``y_true`` values  must be ``0`` and ``1``.
        prefix: The key prefix in the returned dictionary.
        suffix: The key suffix in the returned dictionary.
        nan_policy: The policy on how to handle NaN values in the input
            arrays. The following options are available: ``'omit'``,
            ``'propagate'``, and ``'raise'``.

    Returns:
        The computed metrics.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.metric import fbeta_score
    >>> # auto
    >>> fbeta_score(y_true=np.array([1, 0, 0, 1, 1]), y_pred=np.array([1, 0, 0, 1, 1]))
    {'count': 5, 'f1': 1.0}
    >>> # binary
    >>> fbeta_score(
    ...     y_true=np.array([1, 0, 0, 1, 1]),
    ...     y_pred=np.array([1, 0, 0, 1, 1]),
    ...     label_type="binary",
    ... )
    {'count': 5, 'f1': 1.0}
    >>> # multiclass
    >>> fbeta_score(
    ...     y_true=np.array([0, 0, 1, 1, 2, 2]),
    ...     y_pred=np.array([0, 0, 1, 1, 2, 2]),
    ...     label_type="multiclass",
    ... )
    {'count': 6,
     'f1': array([1., 1., 1.]),
     'macro_f1': 1.0,
     'micro_f1': 1.0,
     'weighted_f1': 1.0}
    >>> # multilabel
    >>> fbeta_score(
    ...     y_true=np.array([[1, 0, 1], [0, 1, 0], [0, 1, 0], [1, 0, 1], [1, 0, 1]]),
    ...     y_pred=np.array([[1, 0, 1], [0, 1, 0], [0, 1, 0], [1, 0, 1], [1, 0, 1]]),
    ...     label_type="multilabel",
    ... )
    {'count': 5,
     'f1': array([1., 1., 1.]),
     'macro_f1': 1.0,
     'micro_f1': 1.0,
     'weighted_f1': 1.0}

    ```
    """
    check_label_type(label_type)
    if label_type == "auto":
        label_type = find_label_type(y_true=y_true, y_pred=y_pred)
    if label_type == "binary":
        return binary_fbeta_score(
            y_true=y_true.ravel(),
            y_pred=y_pred.ravel(),
            betas=betas,
            prefix=prefix,
            suffix=suffix,
            nan_policy=nan_policy,
        )
    if label_type == "multilabel":
        return multilabel_fbeta_score(
            y_true=y_true,
            y_pred=y_pred,
            betas=betas,
            prefix=prefix,
            suffix=suffix,
            nan_policy=nan_policy,
        )
    return multiclass_fbeta_score(
        y_true=y_true.ravel(),
        y_pred=y_pred.ravel(),
        betas=betas,
        prefix=prefix,
        suffix=suffix,
        nan_policy=nan_policy,
    )


def binary_fbeta_score(
    y_true: np.ndarray,
    y_pred: np.ndarray,
    *,
    betas: Sequence[float] = (1,),
    prefix: str = "",
    suffix: str = "",
    nan_policy: str = "propagate",
) -> dict[str, float]:
    r"""Return the F-beta metrics for binary labels.

    Args:
        y_true: The ground truth target labels. This input must
            be an array of shape ``(n_samples,)``.
        y_pred: The predicted labels. This input must
            be an array of shape ``(n_samples,)``.
        betas: The betas used to compute the F-beta scores.
        prefix: The key prefix in the returned dictionary.
        suffix: The key suffix in the returned dictionary.
        nan_policy: The policy on how to handle NaN values in the input
            arrays. The following options are available: ``'omit'``,
            ``'propagate'``, and ``'raise'``.

    Returns:
        The computed metrics.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.metric import binary_fbeta_score
    >>> binary_fbeta_score(
    ...     y_true=np.array([1, 0, 0, 1, 1]),
    ...     y_pred=np.array([1, 0, 0, 1, 1]),
    ... )
    {'count': 5, 'f1': 1.0}

    ```
    """
    y_true, y_pred = preprocess_pred(
        y_true=y_true.ravel(), y_pred=y_pred.ravel(), drop_nan=nan_policy == "omit"
    )
    y_true_nan = contains_nan(arr=y_true, nan_policy=nan_policy, name="'y_true'")
    y_pred_nan = contains_nan(arr=y_pred, nan_policy=nan_policy, name="'y_pred'")

    count = y_true.size
    out = {f"{prefix}count{suffix}": count}
    for beta in betas:
        score = float("nan")
        if count > 0 and not y_true_nan and not y_pred_nan:
            score = float(metrics.fbeta_score(y_true=y_true, y_pred=y_pred, beta=beta))
        out[f"{prefix}f{beta}{suffix}"] = score
    return out


def multiclass_fbeta_score(
    y_true: np.ndarray,
    y_pred: np.ndarray,
    *,
    betas: Sequence[float] = (1,),
    prefix: str = "",
    suffix: str = "",
    nan_policy: str = "propagate",
) -> dict[str, float]:
    r"""Return the F-beta metrics for multiclass labels.

    Args:
        y_true: The ground truth target labels. This input must
            be an array of shape ``(n_samples,)``.
        y_pred: The predicted labels. This input must
            be an array of shape ``(n_samples,)``.
        betas: The betas used to compute the F-beta scores.
        prefix: The key prefix in the returned dictionary.
        suffix: The key suffix in the returned dictionary.
        nan_policy: The policy on how to handle NaN values in the input
            arrays. The following options are available: ``'omit'``,
            ``'propagate'``, and ``'raise'``.

    Returns:
        The computed metrics.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.metric import multiclass_fbeta_score
    >>> multiclass_fbeta_score(
    ...     y_true=np.array([0, 0, 1, 1, 2, 2]),
    ...     y_pred=np.array([0, 0, 1, 1, 2, 2]),
    ... )
    {'count': 6,
     'f1': array([1., 1., 1.]),
     'macro_f1': 1.0,
     'micro_f1': 1.0,
     'weighted_f1': 1.0}

    ```
    """
    y_true, y_pred = preprocess_pred(
        y_true=y_true.ravel(), y_pred=y_pred.ravel(), drop_nan=nan_policy == "omit"
    )
    y_true_nan = contains_nan(arr=y_true, nan_policy=nan_policy, name="'y_true'")
    y_pred_nan = contains_nan(arr=y_pred, nan_policy=nan_policy, name="'y_pred'")

    n_samples = y_true.shape[0]
    out = {f"{prefix}count{suffix}": n_samples}
    for beta in betas:
        fbeta = np.array([])
        macro, micro, weighted = float("nan"), float("nan"), float("nan")
        if n_samples > 0 and not y_true_nan and not y_pred_nan:
            macro = metrics.fbeta_score(
                y_true=y_true, y_pred=y_pred, beta=beta, average="macro", zero_division=0.0
            )
            micro = metrics.fbeta_score(
                y_true=y_true, y_pred=y_pred, beta=beta, average="micro", zero_division=0.0
            )
            weighted = metrics.fbeta_score(
                y_true=y_true, y_pred=y_pred, beta=beta, average="weighted", zero_division=0.0
            )
            fbeta = np.asarray(
                metrics.fbeta_score(
                    y_true=y_true, y_pred=y_pred, beta=beta, average=None, zero_division=0.0
                )
            ).ravel()
        out |= {
            f"{prefix}f{beta}{suffix}": fbeta,
            f"{prefix}macro_f{beta}{suffix}": float(macro),
            f"{prefix}micro_f{beta}{suffix}": float(micro),
            f"{prefix}weighted_f{beta}{suffix}": float(weighted),
        }
    return out


def multilabel_fbeta_score(
    y_true: np.ndarray,
    y_pred: np.ndarray,
    *,
    betas: Sequence[float] = (1,),
    prefix: str = "",
    suffix: str = "",
    nan_policy: str = "propagate",
) -> dict[str, float]:
    r"""Return the F-beta metrics for multilabel labels.

    Args:
        y_true: The ground truth target labels. This input must
            be an array of shape ``(n_samples,)``.
        y_pred: The predicted labels. This input must
            be an array of shape ``(n_samples,)``.
        betas: The betas used to compute the F-beta scores.
        prefix: The key prefix in the returned dictionary.
        suffix: The key suffix in the returned dictionary.
        nan_policy: The policy on how to handle NaN values in the input
            arrays. The following options are available: ``'omit'``,
            ``'propagate'``, and ``'raise'``.

    Returns:
        The computed metrics.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.metric import multilabel_fbeta_score
    >>> multilabel_fbeta_score(
    ...     y_true=np.array([[1, 0, 1], [0, 1, 0], [0, 1, 0], [1, 0, 1], [1, 0, 1]]),
    ...     y_pred=np.array([[1, 0, 1], [0, 1, 0], [0, 1, 0], [1, 0, 1], [1, 0, 1]]),
    ... )
    {'count': 5,
     'f1': array([1., 1., 1.]),
     'macro_f1': 1.0,
     'micro_f1': 1.0,
     'weighted_f1': 1.0}

    ```
    """
    y_true, y_pred = preprocess_pred_multilabel(y_true, y_pred, drop_nan=nan_policy == "omit")
    y_true_nan = contains_nan(arr=y_true, nan_policy=nan_policy, name="'y_true'")
    y_pred_nan = contains_nan(arr=y_pred, nan_policy=nan_policy, name="'y_pred'")

    n_samples = y_true.shape[0]
    out = {f"{prefix}count{suffix}": n_samples}
    for beta in betas:
        fbeta = np.array([])
        macro, micro, weighted = float("nan"), float("nan"), float("nan")
        if n_samples > 0 and not y_true_nan and not y_pred_nan:
            fbeta = np.array(
                metrics.fbeta_score(
                    y_true=y_true,
                    y_pred=y_pred,
                    beta=beta,
                    average="binary" if y_pred.shape[1] == 1 else None,
                )
            ).ravel()
            macro = metrics.fbeta_score(y_true=y_true, y_pred=y_pred, beta=beta, average="macro")
            micro = metrics.fbeta_score(y_true=y_true, y_pred=y_pred, beta=beta, average="micro")
            weighted = metrics.fbeta_score(
                y_true=y_true, y_pred=y_pred, beta=beta, average="weighted"
            )
        out |= {
            f"{prefix}f{beta}{suffix}": fbeta,
            f"{prefix}macro_f{beta}{suffix}": float(macro),
            f"{prefix}micro_f{beta}{suffix}": float(micro),
            f"{prefix}weighted_f{beta}{suffix}": float(weighted),
        }
    return out



r"""Contain a content generator that combines a mapping of content
generators."""

from __future__ import annotations

__all__ = ["ContentGeneratorDict"]

import logging
from typing import TYPE_CHECKING, Any

from coola import objects_are_equal
from coola.utils import repr_indent, repr_mapping, str_indent

from arkas.content.base import BaseContentGenerator
from arkas.utils.html import GO_TO_TOP, render_toc, tags2id, tags2title, valid_h_tag

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence

logger = logging.getLogger(__name__)


class ContentGeneratorDict(BaseContentGenerator):
    r"""Implement a content generator that combines a mapping of content
    generators.

    Args:
        generators: The mapping of content generators.

    Example usage:

    ```pycon

    >>> from arkas.content import ContentGeneratorDict, ContentGenerator
    >>> content = ContentGeneratorDict(
    ...     {
    ...         "one": ContentGenerator(),
    ...         "two": ContentGenerator("meow"),
    ...     }
    ... )
    >>> content
    ContentGeneratorDict(
      (one): ContentGenerator()
      (two): ContentGenerator()
    )

    ```
    """

    def __init__(
        self, generators: Mapping[str, BaseContentGenerator], max_toc_depth: int = 0
    ) -> None:
        self._generators = generators
        self._max_toc_depth = max_toc_depth

    def __repr__(self) -> str:
        args = repr_indent(repr_mapping(self._generators))
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def compute(self) -> ContentGeneratorDict:
        return ContentGeneratorDict(
            generators={key: generator.compute() for key, generator in self._generators.items()},
            max_toc_depth=self._max_toc_depth,
        )

    def equal(self, other: Any, equal_nan: bool = False) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return objects_are_equal(self._generators, other._generators, equal_nan=equal_nan)

    def generate_body(self, number: str = "", tags: Sequence[str] = (), depth: int = 0) -> str:
        report = []
        if tags:
            report.extend(
                [
                    f'<h{valid_h_tag(depth + 1)} id="{tags2id(tags)}">{number} '
                    f"{tags2title(tags)} </h{valid_h_tag(depth + 1)}>",
                    GO_TO_TOP,
                    '<p style="margin-top: 1rem;">',
                ]
            )

        if self._max_toc_depth > 0:
            report.append(
                self._generate_toc_content(
                    number=number, tags=tags, depth=0, max_depth=self._max_toc_depth
                )
            )

        for i, (name, generator) in enumerate(self._generators.items()):
            report.append(
                generator.generate_body(
                    number=f"{number}{i + 1}.", tags=[*list(tags), name], depth=depth + 1
                )
            )
        return "\n".join(report)

    def generate_toc(
        self, number: str = "", tags: Sequence[str] = (), depth: int = 0, max_depth: int = 1
    ) -> str:
        if depth >= max_depth:
            return ""
        toc = []
        if tags:
            toc.append(render_toc(number=number, tags=tags, depth=depth, max_depth=max_depth))
        subtoc = self._generate_toc_content(tags=tags, depth=depth + 1, max_depth=max_depth)
        if subtoc:
            toc.append(subtoc)
        return "\n".join(toc)

    def _generate_toc_content(
        self, number: str = "", tags: Sequence[str] = (), depth: int = 0, max_depth: int = 1
    ) -> str:
        subtoc = []
        for i, (name, generator) in enumerate(self._generators.items()):
            line = generator.generate_toc(
                number=f"{number}{i + 1}.",
                tags=[*list(tags), name],
                depth=depth,
                max_depth=max_depth,
            )
            if line:
                subtoc.append(f"  {str_indent(line)}")
        if subtoc:
            subtoc.insert(0, "<ul>")
            subtoc.append("</ul>")
        return "\n".join(subtoc)



r"""Implement an analyzer that generates a summary of the DataFrame."""

from __future__ import annotations

__all__ = ["DataFrameSummaryAnalyzer"]

import logging
from typing import TYPE_CHECKING

from arkas.analyzer.lazy import BaseLazyAnalyzer
from arkas.output.frame_summary import DataFrameSummaryOutput
from arkas.utils.validation import check_positive

if TYPE_CHECKING:
    import polars as pl

logger = logging.getLogger(__name__)


class DataFrameSummaryAnalyzer(BaseLazyAnalyzer):
    r"""Implement an analyzer to show a summary of the DataFrame.

    Args:
        top: The number of most frequent values to show.
        sort: If ``True``, sort the columns by alphabetical order.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from arkas.analyzer import DataFrameSummaryAnalyzer
    >>> analyzer = DataFrameSummaryAnalyzer()
    >>> analyzer
    DataFrameSummaryAnalyzer(top=5, sort=False)
    >>> frame = pl.DataFrame(
    ...     {
    ...         "col1": [0, 1, 0, 1],
    ...         "col2": [1, 0, 1, 0],
    ...         "col3": [1, 1, 1, 1],
    ...     },
    ...     schema={"col1": pl.Int64, "col2": pl.Int64, "col3": pl.Int64},
    ... )
    >>> output = analyzer.analyze(frame)
    >>> output
    DataFrameSummaryOutput(shape=(4, 3), top=5)

    ```
    """

    def __init__(self, top: int = 5, sort: bool = False) -> None:
        check_positive(name="top", value=top)
        self._top = top
        self._sort = bool(sort)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(top={self._top:,}, sort={self._sort})"

    def _analyze(self, frame: pl.DataFrame) -> DataFrameSummaryOutput:
        logger.info("Analyzing the DataFrame...")
        if self._sort:
            frame = frame.select(sorted(frame.columns))
        return DataFrameSummaryOutput(frame=frame, top=self._top)



r"""Define some utility functions for testing."""

from __future__ import annotations

__all__ = [
    "colorlog_available",
    "hya_available",
    "hydra_available",
    "markdown_available",
    "matplotlib_available",
    "omegaconf_available",
    "plotly_available",
    "scipy_available",
]

import pytest

from arkas.utils.imports import (
    is_colorlog_available,
    is_hya_available,
    is_hydra_available,
    is_markdown_available,
    is_matplotlib_available,
    is_omegaconf_available,
    is_plotly_available,
    is_scipy_available,
)

colorlog_available = pytest.mark.skipif(not is_colorlog_available(), reason="requires colorlog")
hya_available = pytest.mark.skipif(not is_hya_available(), reason="requires hya")
hydra_available = pytest.mark.skipif(not is_hydra_available(), reason="requires hydra")
markdown_available = pytest.mark.skipif(not is_markdown_available(), reason="requires markdown")
matplotlib_available = pytest.mark.skipif(
    not is_matplotlib_available(), reason="requires matplotlib"
)
omegaconf_available = pytest.mark.skipif(not is_omegaconf_available(), reason="requires omegaconf")
plotly_available = pytest.mark.skipif(not is_plotly_available(), reason="requires plotly")
scipy_available = pytest.mark.skipif(not is_scipy_available(), reason="requires scipy")



r"""Contain an exporter that sequentially calls several exporters."""

from __future__ import annotations

__all__ = ["SequentialExporter"]

import logging
from typing import TYPE_CHECKING

from coola.utils import repr_indent, repr_sequence

from arkas.exporter import BaseExporter, setup_exporter

if TYPE_CHECKING:
    from collections.abc import Sequence

    from arkas.output import BaseOutput


logger = logging.getLogger(__name__)


class SequentialExporter(BaseExporter):
    r"""Implement an exporter that sequentially calls several exporters.

    Args:
        exporters: The sequence of exporters.

    Example usage:

    ```pycon


    >>> import tempfile
    >>> from pathlib import Path
    >>> import numpy as np
    >>> from arkas.output import AccuracyOutput
    >>> from arkas.state import AccuracyState
    >>> from arkas.exporter import (
    ...     SequentialExporter,
    ...     FigureExporter,
    ...     MetricExporter,
    ... )
    >>> output = AccuracyOutput(
    ...     state=AccuracyState(
    ...         y_true=np.array([1, 0, 0, 1, 1]),
    ...         y_pred=np.array([1, 0, 0, 1, 1]),
    ...         y_true_name="target",
    ...         y_pred_name="pred",
    ...     )
    ... )
    >>> with tempfile.TemporaryDirectory() as tmpdir:
    ...     path = Path(tmpdir)
    ...     exporter = SequentialExporter(
    ...         [
    ...             MetricExporter(path.joinpath("metrics.pkl")),
    ...             FigureExporter(path.joinpath("figures.pkl")),
    ...         ]
    ...     )
    ...     print(exporter)
    ...     exporter.export(output)
    ...
    SequentialExporter(
      (0): MetricExporter(
          (path): .../metrics.pkl
          (saver): PickleSaver(protocol=5)
          (exist_ok): False
          (show_metrics): True
        )
      (1): FigureExporter(
          (path): .../figures.pkl
          (saver): PickleSaver(protocol=5)
          (exist_ok): False
        )
    )

    ```
    """

    def __init__(self, exporters: Sequence[BaseExporter | dict]) -> None:
        self._exporters = tuple(setup_exporter(exporter) for exporter in exporters)

    def __repr__(self) -> str:
        args = repr_indent(repr_sequence(self._exporters))
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def export(self, output: BaseOutput) -> None:
        for exporter in self._exporters:
            exporter.export(output)



r"""Contain the base class to implement an analyzer."""

from __future__ import annotations

__all__ = ["BaseAnalyzer", "is_analyzer_config", "setup_analyzer"]

import logging
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

from objectory import AbstractFactory
from objectory.utils import is_object_config

if TYPE_CHECKING:
    import polars as pl

    from arkas.output import BaseOutput

logger = logging.getLogger(__name__)


class BaseAnalyzer(ABC, metaclass=AbstractFactory):
    r"""Define the base class to analyze a DataFrame.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from arkas.analyzer import AccuracyAnalyzer
    >>> analyzer = AccuracyAnalyzer(y_true="target", y_pred="pred")
    >>> analyzer
    AccuracyAnalyzer(y_true='target', y_pred='pred', drop_nulls=True, missing_policy='raise', nan_policy='propagate')
    >>> data = pl.DataFrame({"pred": [3, 2, 0, 1, 0, 1], "target": [3, 2, 0, 1, 0, 1]})
    >>> output = analyzer.analyze(data)
    >>> output
    AccuracyOutput(
      (state): AccuracyState(y_true=(6,), y_pred=(6,), y_true_name='target', y_pred_name='pred')
      (nan_policy): propagate
    )

    ```
    """

    @abstractmethod
    def analyze(self, frame: pl.DataFrame, lazy: bool = True) -> BaseOutput:
        r"""Analyze the DataFrame.

        Args:
            frame: The DataFrame to analyze.
            lazy: If ``True``, it forces the computation of the output,
                otherwise it returns an output object that contains the
                logic.

        Returns:
            The generated output.

        Example usage:

        ```pycon

        >>> import polars as pl
        >>> from arkas.analyzer import AccuracyAnalyzer
        >>> analyzer = AccuracyAnalyzer(y_true="target", y_pred="pred")
        >>> data = pl.DataFrame({"pred": [3, 2, 0, 1, 0, 1], "target": [3, 2, 0, 1, 0, 1]})
        >>> output = analyzer.analyze(data)
        >>> output
        AccuracyOutput(
          (state): AccuracyState(y_true=(6,), y_pred=(6,), y_true_name='target', y_pred_name='pred')
          (nan_policy): propagate
        )

        ```
        """


def is_analyzer_config(config: dict) -> bool:
    r"""Indicate if the input configuration is a configuration for a
    ``BaseAnalyzer``.

    This function only checks if the value of the key  ``_target_``
    is valid. It does not check the other values. If ``_target_``
    indicates a function, the returned type hint is used to check
    the class.

    Args:
        config: The configuration to check.

    Returns:
        ``True`` if the input configuration is a configuration
            for a ``BaseAnalyzer`` object.

    Example usage:

    ```pycon

    >>> from arkas.analyzer import is_analyzer_config
    >>> is_analyzer_config({"_target_": "arkas.analyzer.AccuracyAnalyzer"})
    True

    ```
    """
    return is_object_config(config, BaseAnalyzer)


def setup_analyzer(
    analyzer: BaseAnalyzer | dict,
) -> BaseAnalyzer:
    r"""Set up an analyzer.

    The analyzer is instantiated from its configuration
    by using the ``BaseAnalyzer`` factory function.

    Args:
        analyzer: An analyzer or its configuration.

    Returns:
        An instantiated analyzer.

    Example usage:

    ```pycon

    >>> from arkas.analyzer import setup_analyzer
    >>> analyzer = setup_analyzer(
    ...     {
    ...         "_target_": "arkas.analyzer.AccuracyAnalyzer",
    ...         "y_true": "target",
    ...         "y_pred": "pred",
    ...     }
    ... )
    >>> analyzer
    AccuracyAnalyzer(y_true='target', y_pred='pred', drop_nulls=True, missing_policy='raise', nan_policy='propagate')

    ```
    """
    if isinstance(analyzer, dict):
        logger.info("Initializing an analyzer from its configuration... ")
        analyzer = BaseAnalyzer.factory(**analyzer)
    if not isinstance(analyzer, BaseAnalyzer):
        logger.warning(f"analyzer is not a `BaseAnalyzer` (received: {type(analyzer)})")
    return analyzer



r"""Implement the DataFrame summary output."""

from __future__ import annotations

__all__ = ["DataFrameSummaryOutput"]

from typing import TYPE_CHECKING, Any

from coola import objects_are_equal

from arkas.content.frame_summary import DataFrameSummaryContentGenerator
from arkas.evaluator2.vanilla import Evaluator
from arkas.output.lazy import BaseLazyOutput
from arkas.plotter.vanilla import Plotter
from arkas.utils.validation import check_positive

if TYPE_CHECKING:
    import polars as pl


class DataFrameSummaryOutput(BaseLazyOutput):
    r"""Implement the DataFrame summary output.

    Args:
        frame: The DataFrame to analyze.
        top: The number of most frequent values to show.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from arkas.output import DataFrameSummaryOutput
    >>> frame = pl.DataFrame(
    ...     {
    ...         "col1": [1.2, 4.2, 4.2, 2.2],
    ...         "col2": [1, 1, 1, 1],
    ...         "col3": [1, 2, 2, 2],
    ...     },
    ...     schema={"col1": pl.Float64, "col2": pl.Int64, "col3": pl.Int64},
    ... )
    >>> output = DataFrameSummaryOutput(frame)
    >>> output
    DataFrameSummaryOutput(shape=(4, 3), top=5)
    >>> output.get_content_generator()
    DataFrameSummaryContentGenerator(shape=(4, 3), top=5)
    >>> output.get_evaluator()
    Evaluator(count=0)
    >>> output.get_plotter()
    Plotter(count=0)

    ```
    """

    def __init__(self, frame: pl.DataFrame, top: int = 5) -> None:
        self._frame = frame
        check_positive(name="top", value=top)
        self._top = top

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(shape={self._frame.shape}, top={self._top})"

    def equal(self, other: Any, equal_nan: bool = False) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return self._top == other._top and objects_are_equal(
            self._frame, other._frame, equal_nan=equal_nan
        )

    def _get_content_generator(self) -> DataFrameSummaryContentGenerator:
        return DataFrameSummaryContentGenerator(frame=self._frame, top=self._top)

    def _get_evaluator(self) -> Evaluator:
        return Evaluator()

    def _get_plotter(self) -> Plotter:
        return Plotter()



r"""Contain utility functions to draw lines."""

from __future__ import annotations

__all__ = ["axvline_median", "axvline_quantile"]

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from matplotlib.axes import Axes


def axvline_quantile(
    ax: Axes,
    quantile: float,
    label: str,
    color: str = "black",
    linestyle: str = "dashed",
    horizontalalignment: str = "center",
) -> None:
    r"""Add a vertical line to represent a quantile value.

    Args:
        ax: The Axes object that encapsulates all the elements of an
            individual (sub-)plot in a figure.
        quantile: The quantile value.
        label: The associated label to show on the plot.
        color: The line color.
        linestyle: The line style.
        horizontalalignment: The horizontal alignment relative to
            the anchor point.

    Example usage:

    ```pycon

    >>> from matplotlib import pyplot as plt
    >>> from arkas.plot.utils import axvline_quantile
    >>> fig, ax = plt.subplots()
    >>> axvline_quantile(ax, quantile=42.0, label=" q0.9")

    ```
    """
    ax.axvline(quantile, color=color, linestyle=linestyle)
    ax.text(
        quantile,
        0.99,
        label,
        transform=ax.get_xaxis_transform(),
        color=color,
        horizontalalignment=horizontalalignment,
        verticalalignment="top",
    )


def axvline_median(
    ax: Axes,
    median: float,
    label: str = "median",
    color: str = "black",
    linestyle: str = "dashed",
    horizontalalignment: str = "center",
) -> None:
    r"""Add a vertical line to represent the median value.

    Args:
        ax: The Axes object that encapsulates all the elements of an
            individual (sub-)plot in a figure.
        median: The median value.
        label: The associated label to show on the plot.
        color: The line color.
        linestyle: The line style.
        horizontalalignment: The horizontal alignment relative to
            the anchor point.

    Example usage:

    ```pycon

    >>> from matplotlib import pyplot as plt
    >>> from arkas.plot.utils import axvline_median
    >>> fig, ax = plt.subplots()
    >>> axvline_median(ax, median=42.0)

    ```
    """
    ax.axvline(median, color=color, linestyle=linestyle)
    ax.text(
        median,
        0.99,
        label,
        transform=ax.get_xaxis_transform(),
        color=color,
        horizontalalignment=horizontalalignment,
        verticalalignment="top",
    )



r"""Contain text utility functions."""

from __future__ import annotations

__all__ = ["markdown_to_html"]


from arkas.utils.imports import check_markdown, is_markdown_available

if is_markdown_available():  # pragma: no cover
    import markdown


def markdown_to_html(text: str, ignore_error: bool = False) -> str:
    r"""Convert a markdown text to HTML text.

    Args:
        text: The markdown text to convert.
        ignore_error: If ``False``, an error is raised if ``markdown``
            is not installed, otherwise the input text is returned.

    Returns:
        The converted text if ``markdown`` is installed,
            otherwise the input text.

    Example usage:

    ```pycon

    >>> from arkas.utils.text import markdown_to_html
    >>> out = markdown_to_html("- a\n- b\n- c")
    >>> print(out)
    <ul>
    <li>a</li>
    <li>b</li>
    <li>c</li>
    </ul>

    ```
    """
    if not ignore_error:
        check_markdown()
    if is_markdown_available():
        return markdown.markdown(text)
    return text



r"""Contain a simple evaluation runner implementation."""

from __future__ import annotations

__all__ = ["EvaluationRunner"]

import logging
from typing import TYPE_CHECKING, Any

from coola.nested import to_flat_dict
from coola.utils import str_indent, str_mapping
from coola.utils.path import sanitize_path
from grizz.ingestor import BaseIngestor, setup_ingestor
from grizz.transformer import BaseTransformer, setup_transformer
from iden.io import BaseSaver, setup_saver

from arkas.evaluator import BaseEvaluator, setup_evaluator
from arkas.runner.base import BaseRunner

if TYPE_CHECKING:
    from pathlib import Path

logger = logging.getLogger(__name__)


class EvaluationRunner(BaseRunner):
    r"""Implement a simple evaluation runner.

    Args:
        ingestor: The data ingestor or its configuration.
        transformer: The data transformer or its configuration.
        evaluator: The evaluator or its configuration.
        saver: The metric saver or its configuration.
        path: The path where to save the metrics.
        show_metrics: If ``True``, the metrics are shown in the
            logging output.

    Example usage:

    ```pycon

    >>> import tempfile
    >>> import polars as pl
    >>> from pathlib import Path
    >>> from iden.io import PickleSaver
    >>> from grizz.ingestor import Ingestor
    >>> from grizz.transformer import SequentialTransformer
    >>> from arkas.evaluator import AccuracyEvaluator
    >>> from arkas.runner import EvaluationRunner
    >>> with tempfile.TemporaryDirectory() as tmpdir:
    ...     path = Path(tmpdir).joinpath("metrics.pkl")
    ...     runner = EvaluationRunner(
    ...         ingestor=Ingestor(
    ...             pl.DataFrame(
    ...                 {
    ...                     "pred": [3, 2, 0, 1, 0],
    ...                     "target": [3, 2, 0, 1, 0],
    ...                 }
    ...             )
    ...         ),
    ...         transformer=SequentialTransformer(transformers=[]),
    ...         evaluator=AccuracyEvaluator(y_true="target", y_pred="pred"),
    ...         saver=PickleSaver(),
    ...         path=path,
    ...     )
    ...     print(runner)
    ...     runner.run()
    ...
    EvaluationRunner(
      (ingestor): Ingestor(shape=(5, 2))
      (transformer): SequentialTransformer()
      (evaluator): AccuracyEvaluator(y_true='target', y_pred='pred', drop_nulls=True, nan_policy='propagate')
      (saver): PickleSaver(protocol=5)
      (path): .../metrics.pkl
      (show_metrics): True
    )

    ```
    """

    def __init__(
        self,
        ingestor: BaseIngestor | dict,
        transformer: BaseTransformer | dict,
        evaluator: BaseEvaluator | dict,
        saver: BaseSaver | dict,
        path: Path | str,
        show_metrics: bool = True,
    ) -> None:
        self._ingestor = setup_ingestor(ingestor)
        self._transformer = setup_transformer(transformer)
        self._evaluator = setup_evaluator(evaluator)
        self._saver = setup_saver(saver)
        self._path = sanitize_path(path)
        self._show_metrics = bool(show_metrics)

    def __repr__(self) -> str:
        args = str_indent(
            str_mapping(
                {
                    "ingestor": self._ingestor,
                    "transformer": self._transformer,
                    "evaluator": self._evaluator,
                    "saver": self._saver,
                    "path": self._path,
                    "show_metrics": self._show_metrics,
                }
            )
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def run(self) -> Any:
        logger.info("Ingesting data...")
        raw_data = self._ingestor.ingest()
        logger.info("Transforming data...")
        data = self._transformer.transform(raw_data)
        logger.info("Evaluating...")
        result = self._evaluator.evaluate(data)
        logger.info(f"result:\n{result}")
        logger.info("Computing metrics...")
        metrics = result.compute_metrics()
        logger.info(f"Saving metrics at {self._path}...")
        self._saver.save(metrics, path=self._path, exist_ok=True)

        if self._show_metrics:
            logger.info(f"metrics:\n{str_mapping(to_flat_dict(metrics), sorted_keys=True)}")



r"""Contain an plotter that generates figures from a mapping of
plotters."""

from __future__ import annotations

__all__ = ["PlotterDict"]

import logging
from typing import TYPE_CHECKING, Any

from coola import objects_are_equal
from coola.utils import repr_indent, repr_mapping

from arkas.plotter.base import BasePlotter
from arkas.plotter.vanilla import Plotter

if TYPE_CHECKING:
    from collections.abc import Hashable, Mapping

logger = logging.getLogger(__name__)


class PlotterDict(BasePlotter):
    r"""Implement a plotter that generates figures from a mapping of
    plotters.

    Args:
        plotters: The mapping of plotters.

    Example usage:

    ```pycon

    >>> from arkas.plotter import PlotterDict, Plotter
    >>> plotter = PlotterDict(
    ...     {
    ...         "one": Plotter(),
    ...         "two": Plotter({"fig": None}),
    ...     }
    ... )
    >>> plotter
    PlotterDict(
      (one): Plotter(count=0)
      (two): Plotter(count=1)
    )
    >>> figures = plotter.plot()
    >>> figures
    {'one': {}, 'two': {'fig': None}}

    ```
    """

    def __init__(self, plotters: Mapping[Hashable, BasePlotter]) -> None:
        self._plotters = plotters

    def __repr__(self) -> str:
        args = repr_indent(repr_mapping(self._plotters))
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def compute(self) -> Plotter:
        return Plotter(figures=self.plot())

    def equal(self, other: Any, equal_nan: bool = False) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return objects_are_equal(self._plotters, other._plotters, equal_nan=equal_nan)

    def plot(self, prefix: str = "", suffix: str = "") -> dict:
        return {
            key: plotter.plot(prefix=prefix, suffix=suffix)
            for key, plotter in self._plotters.items()
        }



r"""Contain the Pearson correlation evaluator."""

from __future__ import annotations

__all__ = ["PearsonCorrelationEvaluator"]

import logging
from typing import TYPE_CHECKING

from coola.utils.format import repr_mapping_line

from arkas.evaluator.lazy import BaseLazyEvaluator
from arkas.metric.utils import check_nan_policy
from arkas.result import PearsonCorrelationResult, Result
from arkas.utils.array import to_array

if TYPE_CHECKING:
    import polars as pl


logger = logging.getLogger(__name__)


class PearsonCorrelationEvaluator(BaseLazyEvaluator[PearsonCorrelationResult]):
    r"""Implement the Pearson correlation evaluator.

    Args:
        x: The key or column name of the ground truth target
            values.
        y: The key or column name of the predicted values.
        alternative: The alternative hypothesis. Default is 'two-sided'.
            The following options are available:
            - 'two-sided': the correlation is nonzero
            - 'less': the correlation is negative (less than zero)
            - 'greater': the correlation is positive (greater than zero)
        drop_nulls: If ``True``, the rows with null values in
            ``x`` or ``y`` columns are dropped.
        nan_policy: The policy on how to handle NaN values in the input
            arrays. The following options are available: ``'omit'``,
            ``'propagate'``, and ``'raise'``.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from arkas.evaluator import PearsonCorrelationEvaluator
    >>> evaluator = PearsonCorrelationEvaluator(x="target", y="pred")
    >>> evaluator
    PearsonCorrelationEvaluator(x='target', y='pred', alternative='two-sided', drop_nulls=True, nan_policy='propagate')
    >>> data = pl.DataFrame({"pred": [1, 2, 3, 4, 5], "target": [1, 2, 3, 4, 5]})
    >>> result = evaluator.evaluate(data)
    >>> result
    PearsonCorrelationResult(x=(5,), y=(5,), alternative='two-sided', nan_policy='propagate')

    ```
    """

    def __init__(
        self,
        x: str,
        y: str,
        alternative: str = "two-sided",
        drop_nulls: bool = True,
        nan_policy: str = "propagate",
    ) -> None:
        super().__init__(drop_nulls=drop_nulls)
        self._x = x
        self._y = y
        self._alternative = alternative

        check_nan_policy(nan_policy)
        self._nan_policy = nan_policy

    def __repr__(self) -> str:
        args = repr_mapping_line(
            {
                "x": self._x,
                "y": self._y,
                "alternative": self._alternative,
                "drop_nulls": self._drop_nulls,
                "nan_policy": self._nan_policy,
            }
        )
        return f"{self.__class__.__qualname__}({args})"

    def evaluate(self, data: pl.DataFrame, lazy: bool = True) -> PearsonCorrelationResult | Result:
        logger.info(
            f"Evaluating the Pearson correlation | x={self._x} | "
            f"y={self._y} | drop_nulls={self._drop_nulls} | "
            f"nan_policy={self._nan_policy!r}"
        )
        return self._evaluate(data, lazy)

    def _compute_result(self, data: pl.DataFrame) -> PearsonCorrelationResult:
        return PearsonCorrelationResult(
            x=to_array(data[self._x]).ravel(),
            y=to_array(data[self._y]).ravel(),
            alternative=self._alternative,
            nan_policy=self._nan_policy,
        )

    def _get_columns(self) -> tuple[str, ...]:
        return (self._x, self._y)



r"""Contain an analyzer that transforms the data before to analyze
them."""

from __future__ import annotations

__all__ = ["TransformAnalyzer"]

import logging
from typing import TYPE_CHECKING

from coola.utils.format import repr_indent, repr_mapping
from grizz.transformer import BaseTransformer, setup_transformer

from arkas.analyzer.base import BaseAnalyzer, setup_analyzer

if TYPE_CHECKING:
    import polars as pl

    from arkas.output.base import BaseOutput


logger = logging.getLogger(__name__)


class TransformAnalyzer(BaseAnalyzer):
    r"""Implement an analyzer that transforms the data before to analyze
    them.

    Args:
        transformer: The transformer or its configuration.
        analyzer: The analyzer or its configuration.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from arkas.analyzer import AccuracyAnalyzer, TransformAnalyzer
    >>> from grizz.transformer import DropNullRow
    >>> analyzer = TransformAnalyzer(
    ...     transformer=DropNullRow(), analyzer=AccuracyAnalyzer(y_true="target", y_pred="pred")
    ... )
    >>> analyzer
    TransformAnalyzer(
      (transformer): DropNullRowTransformer(columns=None, exclude_columns=(), missing_policy='raise')
      (analyzer): AccuracyAnalyzer(y_true='target', y_pred='pred', drop_nulls=True, missing_policy='raise', nan_policy='propagate')
    )
    >>> frame = pl.DataFrame(
    ...     {"pred": [3, 2, 0, 1, 0, 1, None], "target": [3, 2, 0, 1, 0, 1, None]}
    ... )
    >>> output = analyzer.analyze(frame)
    >>> output
    AccuracyOutput(
      (state): AccuracyState(y_true=(6,), y_pred=(6,), y_true_name='target', y_pred_name='pred')
      (nan_policy): propagate
    )

    ```
    """

    def __init__(
        self,
        transformer: BaseTransformer | dict,
        analyzer: BaseAnalyzer | dict,
    ) -> None:
        self._transformer = setup_transformer(transformer)
        self._analyzer = setup_analyzer(analyzer)

    def __repr__(self) -> str:
        args = repr_indent(
            repr_mapping({"transformer": self._transformer, "analyzer": self._analyzer})
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def analyze(self, frame: pl.DataFrame, lazy: bool = True) -> BaseOutput:
        frame = self._transformer.transform(frame)
        return self._analyzer.analyze(frame=frame, lazy=lazy)



r"""Contain the base class to implement a reporter."""

from __future__ import annotations

__all__ = ["BaseReporter", "is_reporter_config", "setup_reporter"]

import logging
from abc import ABC, abstractmethod

from objectory import AbstractFactory
from objectory.utils import is_object_config

logger = logging.getLogger(__name__)


class BaseReporter(ABC, metaclass=AbstractFactory):
    r"""Define the base class to generate a HTML report.

    Example usage:

    ```pycon

    >>> import tempfile
    >>> from pathlib import Path
    >>> import polars as pl
    >>> from arkas.evaluator import AccuracyEvaluator
    >>> from grizz.ingestor import Ingestor
    >>> from grizz.transformer import SequentialTransformer
    >>> from arkas.reporter import EvalReporter
    >>> with tempfile.TemporaryDirectory() as tmpdir:
    ...     reporter = EvalReporter(
    ...         ingestor=Ingestor(
    ...             pl.DataFrame({"pred": [3, 2, 0, 1, 0, 1], "target": [3, 2, 0, 1, 0, 1]})
    ...         ),
    ...         transformer=SequentialTransformer(transformers=[]),
    ...         evaluator=AccuracyEvaluator(y_true="target", y_pred="pred"),
    ...         report_path=Path(tmpdir).joinpath("report.html"),
    ...     )
    ...     reporter.generate()
    ...

    ```
    """

    @abstractmethod
    def generate(self) -> None:
        r"""Generate a HTML report.

        Example usage:

        ```pycon

        >>> import tempfile
        >>> from pathlib import Path
        >>> import polars as pl
        >>> from arkas.evaluator import AccuracyEvaluator
        >>> from grizz.ingestor import Ingestor
        >>> from grizz.transformer import SequentialTransformer
        >>> from arkas.reporter import EvalReporter
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     reporter = EvalReporter(
        ...         ingestor=Ingestor(
        ...             pl.DataFrame({"pred": [3, 2, 0, 1, 0, 1], "target": [3, 2, 0, 1, 0, 1]})
        ...         ),
        ...         transformer=SequentialTransformer(transformers=[]),
        ...         evaluator=AccuracyEvaluator(y_true="target", y_pred="pred"),
        ...         report_path=Path(tmpdir).joinpath("report.html"),
        ...     )
        ...     reporter.generate()
        ...

        ```
        """


def is_reporter_config(config: dict) -> bool:
    r"""Indicate if the input configuration is a configuration for a
    ``BaseReporter``.

    This function only checks if the value of the key  ``_target_``
    is valid. It does not check the other values. If ``_target_``
    indicates a function, the returned type hint is used to check
    the class.

    Args:
        config: The configuration to check.

    Returns:
        bool: ``True`` if the input configuration is a configuration
            for a ``BaseReporter`` object.

    Example usage:

    ```pycon

    >>> from arkas.reporter import is_reporter_config
    >>> is_reporter_config(
    ...     {
    ...         "_target_": "arkas.reporter.EvalReporter",
    ...         "ingestor": {
    ...             "_target_": "grizz.ingestor.CsvIngestor",
    ...             "path": "/path/to/data.csv",
    ...         },
    ...         "transformer": {"_target_": "grizz.transformer.DropDuplicate"},
    ...         "evaluator": {
    ...             "_target_": "arkas.evaluator.AccuracyEvaluator",
    ...             "y_true": "target",
    ...             "y_pred": "pred",
    ...         },
    ...         "report_path": "/path/to/report.html",
    ...     }
    ... )
    True

    ```
    """
    return is_object_config(config, BaseReporter)


def setup_reporter(
    reporter: BaseReporter | dict,
) -> BaseReporter:
    r"""Set up a reporter.

    The reporter is instantiated from its configuration
    by using the ``BaseReporter`` factory function.

    Args:
        reporter: A reporter or its configuration.

    Returns:
        An instantiated reporter.

    Example usage:

    ```pycon

    >>> from arkas.reporter import setup_reporter
    >>> reporter = setup_reporter(
    ...     {
    ...         "_target_": "arkas.reporter.EvalReporter",
    ...         "ingestor": {
    ...             "_target_": "grizz.ingestor.CsvIngestor",
    ...             "path": "/path/to/data.csv",
    ...         },
    ...         "transformer": {"_target_": "grizz.transformer.DropDuplicate"},
    ...         "evaluator": {
    ...             "_target_": "arkas.evaluator.AccuracyEvaluator",
    ...             "y_true": "target",
    ...             "y_pred": "pred",
    ...         },
    ...         "report_path": "/path/to/report.html",
    ...     }
    ... )
    >>> reporter
    EvalReporter(
      (ingestor): CsvIngestor(path=/path/to/data.csv)
      (transformer): DropDuplicateTransformer(columns=None, exclude_columns=(), missing_policy='raise')
      (evaluator): AccuracyEvaluator(y_true='target', y_pred='pred', drop_nulls=True, nan_policy='propagate')
      (report_path): /path/to/report.html
      (max_toc_depth): 6
    )

    ```
    """
    if isinstance(reporter, dict):
        logger.info("Initializing a reporter from its configuration... ")
        reporter = BaseReporter.factory(**reporter)
    if not isinstance(reporter, BaseReporter):
        logger.warning(f"reporter is not a `BaseReporter` (received: {type(reporter)})")
    return reporter



r"""Implement the Kullback-Leibler (KL) divergence between two 1D
distributions result."""

from __future__ import annotations

__all__ = ["KLDivResult"]

from typing import TYPE_CHECKING, Any

from coola import objects_are_equal

from arkas.metric.distribution.kl import kl_div
from arkas.metric.utils import check_same_shape
from arkas.result.base import BaseResult

if TYPE_CHECKING:
    import numpy as np


class KLDivResult(BaseResult):
    r"""Implement the Kullback-Leibler (KL) divergence between two 1D
    distributions result.

    Args:
        p: The true probability distribution.
        q: The model probability distribution.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.result import KLDivResult
    >>> result = KLDivResult(p=np.array([0.1, 0.6, 0.1, 0.2]), q=np.array([0.2, 0.5, 0.2, 0.1]))
    >>> result
    KLDivResult(p=(4,), q=(4,))
    >>> result.compute_metrics()
    {'size': 4, 'kl_pq': 0.109..., 'kl_qp': 0.116...}

    ```
    """

    def __init__(self, p: np.ndarray, q: np.ndarray) -> None:
        self._p = p.ravel()
        self._q = q.ravel()

        check_same_shape([self._p, self._q])

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(p={self._p.shape}, q={self._q.shape})"

    @property
    def p(self) -> np.ndarray:
        return self._p

    @property
    def q(self) -> np.ndarray:
        return self._q

    def compute_metrics(self, prefix: str = "", suffix: str = "") -> dict[str, float]:
        return kl_div(p=self._p, q=self._q, prefix=prefix, suffix=suffix)

    def equal(self, other: Any, equal_nan: bool = False) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return objects_are_equal(self.p, other.p, equal_nan=equal_nan) and objects_are_equal(
            self.q, other.q, equal_nan=equal_nan
        )

    def generate_figures(
        self, prefix: str = "", suffix: str = ""  # noqa: ARG002
    ) -> dict[str, float]:
        return {}



r"""Implement the Spearman correlation result."""

from __future__ import annotations

__all__ = ["SpearmanCorrelationResult"]

from typing import TYPE_CHECKING, Any

from coola import objects_are_equal
from coola.utils.format import repr_mapping_line

from arkas.metric.correlation.spearman import spearmanr
from arkas.metric.utils import check_nan_policy, check_same_shape
from arkas.result.base import BaseResult

if TYPE_CHECKING:
    import numpy as np


class SpearmanCorrelationResult(BaseResult):
    r"""Implement the Spearman correlation result.

    Args:
        x: The first input array.
        y: The second input array.
        alternative: The alternative hypothesis. Default is 'two-sided'.
            The following options are available:
            - 'two-sided': the correlation is nonzero
            - 'less': the correlation is negative (less than zero)
            - 'greater': the correlation is positive (greater than zero)
        nan_policy: The policy on how to handle NaN values in the input
            arrays. The following options are available: ``'omit'``,
            ``'propagate'``, and ``'raise'``.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.result import SpearmanCorrelationResult
    >>> result = SpearmanCorrelationResult(
    ...     x=np.array([1, 2, 3, 4, 5, 6, 7, 8, 9]),
    ...     y=np.array([1, 2, 3, 4, 5, 6, 7, 8, 9]),
    ... )
    >>> result
    SpearmanCorrelationResult(x=(9,), y=(9,), alternative='two-sided', nan_policy='propagate')
    >>> result.compute_metrics()
    {'count': 9, 'spearman_coeff': 1.0, 'spearman_pvalue': 0.0}

    ```
    """

    def __init__(
        self,
        x: np.ndarray,
        y: np.ndarray,
        alternative: str = "two-sided",
        nan_policy: str = "propagate",
    ) -> None:
        self._x = x.ravel()
        self._y = y.ravel()
        self._alternative = alternative

        check_same_shape([self._x, self._y])

        check_nan_policy(nan_policy)
        self._nan_policy = nan_policy

    def __repr__(self) -> str:
        args = repr_mapping_line(
            {
                "x": self._x.shape,
                "y": self._y.shape,
                "alternative": self._alternative,
                "nan_policy": self._nan_policy,
            }
        )
        return f"{self.__class__.__qualname__}({args})"

    @property
    def nan_policy(self) -> str:
        return self._nan_policy

    @property
    def x(self) -> np.ndarray:
        return self._x

    @property
    def y(self) -> np.ndarray:
        return self._y

    @property
    def alternative(self) -> str:
        return self._alternative

    def compute_metrics(self, prefix: str = "", suffix: str = "") -> dict[str, float]:
        return spearmanr(
            x=self._x,
            y=self._y,
            alternative=self._alternative,
            prefix=prefix,
            suffix=suffix,
            nan_policy=self._nan_policy,
        )

    def equal(self, other: Any, equal_nan: bool = False) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return (
            objects_are_equal(self.x, other.x, equal_nan=equal_nan)
            and objects_are_equal(self.y, other.y, equal_nan=equal_nan)
            and self.alternative == other.alternative
            and self.nan_policy == other.nan_policy
        )

    def generate_figures(
        self, prefix: str = "", suffix: str = ""  # noqa: ARG002
    ) -> dict[str, float]:
        return {}



r"""Implement the simple result."""

from __future__ import annotations

__all__ = ["EmptyResult", "Result"]

from typing import Any

from coola import objects_are_equal

from arkas.result.base import BaseResult


class Result(BaseResult):
    r"""Implement a simple result.

    Args:
        metrics: The metrics.
        figures: The figures.

    Example usage:

    ```pycon

    >>> from arkas.result import Result
    >>> result = Result(metrics={"accuracy": 1.0, "count": 42}, figures={})
    >>> result
    Result(metrics=2, figures=0)
    >>> result.compute_metrics()
    {'accuracy': 1.0, 'count': 42}

    ```
    """

    def __init__(self, metrics: dict | None = None, figures: dict | None = None) -> None:
        self._metrics = metrics or {}
        self._figures = figures or {}

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(metrics={len(self._metrics):,}, figures={len(self._figures):,})"

    def compute_metrics(self, prefix: str = "", suffix: str = "") -> dict:
        return {f"{prefix}{key}{suffix}": value for key, value in self._metrics.items()}

    def equal(self, other: Any, equal_nan: bool = False) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return objects_are_equal(
            self._metrics, other._metrics, equal_nan=equal_nan
        ) and objects_are_equal(self._figures, other._figures, equal_nan=equal_nan)

    def generate_figures(self, prefix: str = "", suffix: str = "") -> dict:
        return {f"{prefix}{key}{suffix}": value for key, value in self._figures.items()}


class EmptyResult(Result):
    r"""Implement an empty result.

    This result is designed to be used when it is possible to evaluate a
    result.
    """

    def __init__(self) -> None:
        super().__init__(metrics={}, figures={})

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}()"



r"""Implement an empty output."""

from __future__ import annotations

__all__ = ["EmptyOutput"]


from arkas.content.vanilla import ContentGenerator
from arkas.evaluator2.vanilla import Evaluator
from arkas.output.vanilla import Output
from arkas.plotter.vanilla import Plotter


class EmptyOutput(Output):
    r"""Implement the accuracy output.

    Example usage:

    ```pycon

    >>> from arkas.output import EmptyOutput
    >>> output = EmptyOutput()
    >>> output
    EmptyOutput()
    >>> output.get_evaluator()
    Evaluator(count=0)
    >>> output.get_plotter()
    Plotter(count=0)

    ```
    """

    def __init__(self) -> None:
        super().__init__(content=ContentGenerator(), evaluator=Evaluator(), plotter=Plotter())

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}()"
